import numpy as np
import torch

import meddlr.ops.complex as cplx


def add_even_odd_motion(
    kspace: torch.Tensor,
    scale: float,
    channel_first: bool = False,
    seed=None,
    generator: torch.Generator = None,
):
    assert cplx.is_complex(kspace) or cplx.is_complex_as_real(kspace)

    phase_matrix = torch.zeros(kspace.shape, dtype=torch.complex64, device=kspace.device)
    width = kspace.shape[3] if channel_first else kspace.shape[2]
    g = generator if seed is None else torch.Generator(device=kspace.device).manual_seed(seed)

    odd_err = (2 * np.pi * scale) * torch.rand(
        1, generator=g, device=g.device
    ).cpu().numpy() - np.pi * scale
    even_err = (2 * np.pi * scale) * torch.rand(
        1, generator=g, device=g.device
    ).cpu().numpy() - np.pi * scale
    for line in range(width):
        if line % 2 == 0:
            rand_err = even_err
        else:
            rand_err = odd_err
        phase_error = torch.from_numpy(np.exp(-1j * rand_err)).item()
        if channel_first:
            phase_matrix[:, :, :, line] = phase_error
        else:
            phase_matrix[:, :, line] = phase_error
    aug_kspace = kspace * phase_matrix
    return aug_kspace



from typing import Union

import torch


class TransformCacheMixin:
    def _reset_transform(self):
        self._transform = None

    def cached_transform(self):
        return self._transform


class DeviceMixin:
    device: Union[str, int, torch.device]

    def to(self, device):
        self._device = device

    def cpu(self):
        return self.to(torch.device("cpu"))

    def cuda(self):
        return self.to(torch.device("cuda"))


class GeometricMixin:
    def is_geometric(self):
        return True



from meddlr.modeling.loss_computer import (  # noqa: F401
    LOSS_COMPUTER_REGISTRY,
    BasicLossComputer,
    N2RLossComputer,
    build_loss_computer,
)
from meddlr.modeling.meta_arch import (  # noqa: F401
    META_ARCH_REGISTRY,
    build_model,
    initialize_model,
)

__all__ = [
    "BasicLossComputer",
    "N2RLossComputer",
    "build_loss_computer",
    "build_model",
    "initialize_model",
    "META_ARCH_REGISTRY",
    "LOSS_COMPUTER_REGISTRY",
]



"""
Implementations of different CNNs.
"""

from torch import nn

from meddlr.utils.transforms import center_crop


class SeparableConv3d(nn.Module):
    """
    A separable 3D convolutional operator.
    """

    def __init__(self, in_chans, out_chans, kernel_size, spatial_chans=None, act_type="relu"):
        """
        Args:
            in_chans (int): Number of channels in the input.
            out_chans (int): Number of channels in the output.
            kernel_size (int): Size of kernel (repeated for all three dimensions).
        """
        super().__init__()

        sp_kernel_size = (1, kernel_size, kernel_size)
        sp_pad_size = (0, 1, 1)
        t_kernel_size = (kernel_size, 1, 1)
        t_pad_size = (1, 0, 0)

        if spatial_chans is None:
            # Force number of spatial features, such that the total number of
            # parameters is the same as a nn.Conv3D(in_chans, out_chans)
            spatial_chans = (kernel_size ** 3) * in_chans * out_chans
            spatial_chans /= (kernel_size ** 2) * in_chans + kernel_size * out_chans
            spatial_chans = int(spatial_chans)

        # Define each layer in SeparableConv3d block
        spatial_conv = nn.Conv3d(
            in_chans, spatial_chans, kernel_size=sp_kernel_size, padding=sp_pad_size
        )
        temporal_conv = nn.Conv3d(
            spatial_chans, out_chans, kernel_size=t_kernel_size, padding=t_pad_size
        )

        # Define choices for intermediate activation layer
        activations = nn.ModuleDict([["none", nn.Identity()]["relu", nn.ReLU()]])

        # Define the forward pass
        self.layers = nn.Sequential(spatial_conv, activations[act_type], temporal_conv)

    def forward(self, input):
        """
        Args:
            input (torch.Tensor): Input tensor of shape ``(B,C_{in},D,H,W)``.

        Returns:
            (torch.Tensor): Output tensor of shape ``(B,C_{in},D,H,W)``.
        """
        return self.layers(input)


class ConvBlock(nn.Module):
    """
    A 3D Convolutional Block that consists of Norm -> ReLU -> Dropout -> Conv

    Based on implementation described by:
        K He, et al. "Identity Mappings in Deep Residual Networks" arXiv:1603.05027
    """

    def __init__(
        self,
        in_chans,
        out_chans,
        kernel_size,
        drop_prob,
        conv_type="conv3d",
        act_type="relu",
        norm_type="none",
    ):
        """
        Args:
            in_chans (int): Number of channels in the input.
            out_chans (int): Number of channels in the output.
            drop_prob (float): Dropout probability.
        """
        super().__init__()

        self.in_chans = in_chans
        self.out_chans = out_chans
        self.drop_prob = drop_prob

        # Define choices for each layer in ConvBlock
        normalizations = nn.ModuleDict(
            [
                ["none", nn.Identity()],
                ["instance", nn.InstanceNorm3d(in_chans, affine=False)],
                ["batch", nn.BatchNorm3d(in_chans, affine=False)],
            ]
        )
        activations = nn.ModuleDict([["relu", nn.ReLU()], ["leaky_relu", nn.LeakyReLU()]])
        dropout = nn.Dropout3d(p=drop_prob, inplace=True)

        # Note: don't use ModuleDict here. Otherwise, the parameters for the un-selected
        # convolution type will still be initialized and added to model.parameters()
        if conv_type == "conv3d":
            convolution = nn.Conv3d(in_chans, out_chans, kernel_size=kernel_size, padding=1)
        else:
            convolution = SeparableConv3d(in_chans, out_chans, kernel_size=kernel_size, padding=1)

        # Define forward pass
        self.layers = nn.Sequential(
            normalizations[norm_type], activations[act_type], dropout, convolution
        )

    def forward(self, input):
        """
        Args:
            input (torch.Tensor): Input tensor of shape ``(B,C_{in},D,H,W)``.

        Returns:
            (torch.Tensor): Output tensor of shape ``(B,C_{in},D,H,W)``.
        """
        return self.layers(input)

    def __repr__(self):
        return (
            f"ConvBlock3D(in_chans={self.in_chans}, out_chans={self.out_chans}, "
            f"drop_prob={self.drop_prob})"
        )


class ResBlock(nn.Module):
    """
    A ResNet block that consists of two convolutional layers followed by a residual connection.
    """

    def __init__(self, in_chans, out_chans, kernel_size, drop_prob):
        """
        Args:
            in_chans (int): Number of channels in the input.
            out_chans (int): Number of channels in the output.
            drop_prob (float): Dropout probability.
        """
        super().__init__()

        self.layers = nn.Sequential(
            ConvBlock(in_chans, out_chans, kernel_size, drop_prob),
            ConvBlock(out_chans, out_chans, kernel_size, drop_prob),
        )

        if in_chans != out_chans:
            self.resample = nn.Conv3d(in_chans, out_chans, kernel_size=1)
        else:
            self.resample = nn.Identity()

    def forward(self, input):
        """
        Args:
            input (torch.Tensor): Input tensor of shape ``(B,C_{in},D,H,W)``.

        Returns:
            (torch.Tensor): Output tensor of shape ``(B,C_{in},D,H,W)``.
        """

        # To have a residual connection, number of inputs must be equal to outputs
        shortcut = self.resample(input)

        return self.layers(input) + shortcut


class ResNet(nn.Module):
    """
    Prototype for 3D ResNet architecture
    """

    def __init__(self, num_resblocks, in_chans, chans, kernel_size, drop_prob, circular_pad=True):
        """ """
        super().__init__()

        self.circular_pad = circular_pad
        self.pad_size = 2 * num_resblocks + 1

        # Declare ResBlock layers
        self.res_blocks = nn.ModuleList([ResBlock(in_chans, chans, kernel_size, drop_prob)])
        for _ in range(num_resblocks - 1):
            self.res_blocks += [ResBlock(chans, chans, kernel_size, drop_prob)]

        # Declare final conv layer (down-sample to original in_chans)
        self.final_layer = nn.Conv3d(chans, in_chans, kernel_size=kernel_size, padding=1)

    def forward(self, input):
        """
        Args:
            input (torch.Tensor): Input tensor of shape ``(B,C_{in},D,H,W)``.

        Returns:
            (torch.Tensor): Output tensor of shape ``(B,C_{in},D,H,W)``.
        """

        orig_shape = input.shape
        if self.circular_pad:
            input = nn.functional.pad(input, (0, 0, 0, 0) + 2 * (self.pad_size,), mode="circular")
            # input = nn.functional.pad(input, 4*(self.pad_size,) + (0,0), mode='replicate')

        # Perform forward pass through the network
        output = input
        for res_block in self.res_blocks:
            output = res_block(output)
        output = self.final_layer(output) + input

        return center_crop(output, orig_shape)



from .build import META_ARCH_REGISTRY, build_model, initialize_model  # noqa: F401
from .cs_model import CSModel  # noqa: F401
from .denoising import DenoisingModel  # noqa: F401
from .generalized_unet import GeneralizedUNet  # noqa: F401
from .m2r import M2RModel  # noqa: F401
from .n2r import N2RModel  # noqa: F401
from .nm2r import NM2RModel  # noqa: F401
from .ssdu import SSDUModel  # noqa: F401
from .unet import UnetModel  # noqa: F401
from .unrolled import GeneralizedUnrolledCNN  # noqa: F401
from .vortex import VortexModel  # noqa: F401



import logging
from typing import Sequence, Tuple

import numpy as np
import torch
import torch.nn.functional as F
import torchvision.transforms.functional as TF
from PIL import Image

import meddlr.ops.complex as cplx
from meddlr.transforms.build import TRANSFORM_REGISTRY
from meddlr.transforms.mixins import GeometricMixin
from meddlr.transforms.transform import Transform


@TRANSFORM_REGISTRY.register()
class AffineTransform(GeometricMixin, Transform):
    def __init__(
        self,
        angle: float = None,
        translate: Sequence[int] = None,
        scale=None,
        shear: Sequence[int] = None,
        pad_like: str = None,
        upsample_factor: float = 1,
        upsample_order: int = 1,
    ) -> None:
        super().__init__()
        logger = logging.getLogger(f"{__name__}.{type(self).__name__}")

        if angle is None:
            angle = 0.0
        if scale is None:
            scale = 1.0
        if translate is None:
            translate = [0, 0]
        if shear is None:
            shear = [0, 0]

        if pad_like not in (None, "MRAugment"):
            raise ValueError("`pad_like` must be one of (None, 'MRAugment')")
        if pad_like == "MRAugment" and translate not in ([0, 0], None):
            logger.warning("MRAugment padding may not appropriately account for translation")
        self._set_attributes(locals())

    def _apply_affine(self, x):
        img = x
        angle = self.angle
        translate = self.translate[::-1]
        scale = self.scale
        shear = self.shear[::-1]
        upsample_factor = self.upsample_factor
        upsample_order = self.upsample_order

        is_complex = cplx.is_complex(img)
        permute = is_complex or cplx.is_complex_as_real(img)
        if is_complex:
            img = torch.view_as_real(img)
        if permute:
            img = img.permute((img.ndim - 1,) + tuple(range(0, img.ndim - 1)))

        shape = img.shape
        use_view = img.ndim > 4
        if use_view:
            is_contiguous = img.is_contiguous()
            if is_contiguous:
                img = img.view((np.product(shape[:-3]),) + shape[-3:])
            else:
                img = img.reshape((np.product(shape[:-3]),) + shape[-3:])

        base_shape = img.shape[-2:]
        upsample = upsample_factor != 1
        interpolation = Image.BICUBIC if upsample_order == 3 else Image.BILINEAR
        if upsample:
            upsampled_shape = (
                img.shape[-2] * self.upsample_factor,
                img.shape[-1] * self.upsample_factor,
            )
            img = TF.resize(img, size=upsampled_shape, interpolation=interpolation)

        h, w = img.shape[-2:]
        if self.pad_like == "MRAugment":
            pad = _get_mraugment_affine_pad(shape[-2:], angle, translate, scale, shear)
            img = TF.pad(img, padding=pad, padding_mode="reflect")

        img = TF.affine(
            img, angle=angle, translate=translate, scale=scale, shear=shear, resample=2  # bilinear
        )

        if self.pad_like == "MRAugment":
            img = TF.center_crop(img, (h, w))

        if upsample:
            img = TF.resize(img, size=base_shape, interpolation=interpolation)

        if use_view:
            img = img.view(shape)

        if permute:
            img = img.permute(tuple(range(1, img.ndim)) + (0,))
        if is_complex:
            img = torch.view_as_complex(img.contiguous())
        return img

    def apply_image(self, img: torch.Tensor):
        return self._apply_affine(img)

    def apply_maps(self, maps: torch.Tensor):
        maps = self._apply_affine(maps)  # BxCxMxHxW
        norm = cplx.rss(maps, dim=1).unsqueeze(1)
        norm += 1e-8 * (norm == 0)
        maps = maps / norm
        return maps

    def _eq_attrs(self) -> Tuple[str]:
        return ("angle", "translate", "scale", "shear")


@TRANSFORM_REGISTRY.register()
class TranslationTransform(GeometricMixin, Transform):
    def __init__(self, translate: Sequence[int], pad_mode="constant", pad_value=0) -> None:
        super().__init__()
        self.translate = translate
        self.pad_mode = pad_mode
        self.pad_value = pad_value

    def apply_image(self, img: torch.Tensor):
        translation = self.translate
        max_dims = len(translation) + 2

        is_complex = cplx.is_complex(img)
        permute = is_complex or cplx.is_complex_as_real(img)
        if is_complex:
            img = torch.view_as_real(img)
        if permute:
            img = img.permute((img.ndim - 1,) + tuple(range(0, img.ndim - 1)))

        shape = img.shape
        use_view = img.ndim > max_dims
        if use_view:
            is_contiguous = img.is_contiguous()
            if is_contiguous:
                img = img.view((np.product(shape[: -(max_dims - 1)]),) + shape[-(max_dims - 1) :])
            else:
                img = img.reshape(
                    (np.product(shape[: -(max_dims - 1)]),) + shape[-(max_dims - 1) :]
                )

        pad, sl = _get_mraugment_translate_pad(img.shape, translation)
        img = F.pad(img, pad, mode=self.pad_mode, value=self.pad_value)
        img = img[sl]

        if use_view:
            img = img.view(shape)

        if permute:
            img = img.permute(tuple(range(1, img.ndim)) + (0,))
        if is_complex:
            img = torch.view_as_complex(img.contiguous())
        return img

    def apply_maps(self, maps: torch.Tensor):
        maps = self.apply_image(maps)  # BxCxMxHxW
        norm = cplx.rss(maps, dim=1).unsqueeze(1)
        norm += 1e-8 * (norm == 0)
        maps = maps / norm
        return maps


@TRANSFORM_REGISTRY.register()
class FlipTransform(GeometricMixin, Transform):
    def __init__(self, dims):
        super().__init__()
        if isinstance(dims, int):
            dims = (dims,)
        self.dims = dims

    def apply_image(self, img: torch.Tensor):
        if cplx.is_complex_as_real(img):
            img = torch.view_as_complex(img)
        return torch.flip(img, self.dims)

    def apply_kspace(self, kspace):
        return self.apply_image(kspace)

    def inverse(self):
        return FlipTransform(self.dims)

    def _eq_attrs(self) -> Tuple[str]:
        return ("dims",)


@TRANSFORM_REGISTRY.register()
class Rot90Transform(GeometricMixin, Transform):
    def __init__(self, k, dims) -> None:
        super().__init__()
        self.k = k
        self.dims = dims

    def apply_image(self, img: torch.Tensor):
        return torch.rot90(img, self.k, self.dims)

    def apply_kspace(self, kspace):
        return self.apply_image(kspace)

    def inverse(self):
        return Rot90Transform(self.k, self.dims[::-1])

    def _eq_attrs(self) -> Tuple[str]:
        return ("k", "dims")


def _get_mraugment_affine_pad(im_shape, angle, translate, scale, shear):
    """Calculate the padding size based on MRAugment padding method.

    This padding should be applied before the affine transformation.

    Args:
        im_shape (tuple): Shape as ``(height, width)``.
        angle (float): The rotating angle.
        scale (float): The scale factor.
        shear (tuple): Shear factors (H x W) (i.e. YxX).

    Note:
        This method is adapted from MRAugment.
        https://github.com/MathFLDS/MRAugment/blob/master/mraugment/data_augment.py
    """
    h, w = im_shape
    shear = shear[::-1]
    translate = translate[::-1]

    corners = [
        [-h / 2, -w / 2, 1.0],
        [-h / 2, w / 2, 1.0],
        [h / 2, w / 2, 1.0],
        [h / 2, -w / 2, 1.0],
    ]
    mx = torch.tensor(
        TF._get_inverse_affine_matrix([0.0, 0.0], -angle, translate, scale, [-s for s in shear])
    ).reshape(2, 3)
    corners = torch.cat([torch.tensor(c).reshape(3, 1) for c in corners], dim=1)
    tr_corners = torch.matmul(mx, corners)
    all_corners = torch.cat([tr_corners, corners[:2, :]], dim=1)
    bounding_box = all_corners.amax(dim=1) - all_corners.amin(dim=1)
    py = torch.clip(torch.floor((bounding_box[0] - h) / 2), min=0.0, max=h - 1)
    px = torch.clip(torch.floor((bounding_box[1] - w) / 2), min=0.0, max=w - 1)
    return int(px.item()), int(py.item())


def _get_mraugment_translate_pad(im_shape, translation):
    shape = im_shape[-len(translation) :]

    pad = []
    sl = []

    for s, t in zip(shape, translation):
        if t > 0:
            pad.append((t, 0))
            sl.append(slice(0, s))
        else:
            pad.append((0, abs(t)))
            sl.append(slice(abs(t), None))
    pad = [x for y in pad[::-1] for x in y]
    sl.insert(0, Ellipsis)
    return pad, sl



from typing import Sequence

import torch
import torch.nn.functional as F

import meddlr.ops.complex as cplx

__all__ = [
    "roll",
    "pad",
    "zero_pad",
    "time_average",
    "sliding_window",
    "center_crop",
    "normalize",
    "normalize_instance",
]


def roll(x, shift, dim):
    """
    Similar to np.roll but applies to PyTorch Tensors
    """
    if isinstance(shift, (tuple, list)):
        assert len(shift) == len(dim)
        for s, d in zip(shift, dim):
            x = roll(x, s, d)
        return x
    shift = shift % x.size(dim)
    if shift == 0:
        return x
    left = x.narrow(dim, 0, x.size(dim) - shift)
    right = x.narrow(dim, x.size(dim) - shift, shift)
    return torch.cat((right, left), dim=dim)


def pad(x: torch.Tensor, shape: Sequence[int], mode="constant", value=0):
    """
    Args:
        x: Input tensor of shape (B, ...)
        shape: Shape to zero pad to. Use `None` to skip padding certain dimensions.
    Returns:
    """
    x_shape = x.shape[1 : 1 + len(shape)]
    assert all(
        x_shape[i] <= shape[i] or shape[i] is None for i in range(len(shape))
    ), f"Tensor spatial dimensions {x_shape} smaller than zero pad dimensions"

    total_padding = tuple(
        desired - current if desired is not None else 0 for current, desired in zip(x_shape, shape)
    )
    # Adding no padding for terminal dimensions.
    # torch.nn.functional.pad pads dimensions in reverse order.
    total_padding += (0,) * (len(x.shape) - 1 - len(x_shape))
    total_padding = total_padding[::-1]

    pad = []
    for padding in total_padding:
        pad1 = padding // 2
        pad2 = padding - pad1
        pad.extend([pad1, pad2])

    return F.pad(x, pad, mode=mode, value=value)


def zero_pad(x: torch.Tensor, shape: Sequence[int]):
    return pad(x, shape, mode="constant", value=0)


def time_average(data, dim, eps=1e-6, keepdim=True):
    """
    Computes time average across a specified axis.
    """
    mask = cplx.get_mask(data)
    return data.sum(dim, keepdim=keepdim) / (mask.sum(dim, keepdim=keepdim) + eps)


def sliding_window(data, dim, window_size):
    """
    Computes sliding window with circular boundary conditions across a specified
    axis.
    """
    assert 0 < window_size <= data.shape[dim]

    windows = [None] * data.shape[dim]
    for i in range(data.shape[dim]):
        data_slide = roll(data, int(window_size / 2) - i, dim)
        window = data_slide.narrow(dim, 0, window_size)
        windows[i] = time_average(window, dim)

    return torch.cat(windows, dim=dim)


def center_crop(data, shape):
    """
    Apply a center crop to a batch of images.
    Args:
        data (torch.Tensor): The input tensor to be center cropped.
        shape (list of ints): The output shape. If shape[dim] = -1, then no crop
            will be applied in that dimension.
    """
    for i in range(len(shape)):
        if (shape[i] == data.shape[i]) or (shape[i] == -1):
            continue
        assert 0 < shape[i] <= data.shape[i]
        idx_start = (data.shape[i] - shape[i]) // 2
        data = data.narrow(i, idx_start, shape[i])

    return data


def normalize(data, mean, stddev, eps=0.0):
    """
    Normalize the given tensor using:
        (data - mean) / (stddev + eps)
    Args:
        data (torch.Tensor): Input data to be normalized
        mean (float): Mean value
        stddev (float): Standard deviation
        eps (float): Added to stddev to prevent dividing by zero
    Returns:
        torch.Tensor: Normalized tensor
    """
    return (data - mean) / (stddev + eps)


def normalize_instance(data, eps=0.0):
    """
    Normalize the given tensor using:
        (data - mean) / (stddev + eps)
    where mean and stddev are computed from the data itself.
    Args:
        data (torch.Tensor): Input data to be normalized
        eps (float): Added to stddev to prevent dividing by zero
    Returns:
        torch.Tensor: Normalized tensor
    """
    mean = data.mean()
    std = data.std()
    return normalize(data, mean, std, eps), mean, std



import logging
from collections import defaultdict
from typing import Any, Dict, Hashable, List, Mapping, Sequence, Tuple, Union

import numpy as np
import torch
from torch.utils.data import Sampler, SubsetRandomSampler

from meddlr.utils import comm

_UNKNOWN_TOKEN = "<UNK>"

__all__ = ["GroupSampler", "AlternatingGroupSampler", "DistributedGroupSampler"]


class GroupSampler(Sampler):
    """
    TODO:
        - uniform group sampling (n groups per batch)
        - weighted group sampling (n groups per batch)
        - frequency sampling (duty cycle of each group)
    """

    def __init__(
        self,
        dataset,
        batch_by: Any = None,
        batch_size: int = None,
        as_batch_sampler: bool = False,
        drop_last: bool = False,
        shuffle: bool = False,
        seed: int = None,
    ):
        """
        Args:
            dataset: dataset to sample from.
                Must have attribute ``.examples``, which returns a list of
                dictionaries of examples with associated metadata.
            batch_by (Any, Sequence[Any]): Metadata field(s) to batch by.
            batch_size (int, optional): The batch size. Required if ``batched_by``
                specified or ``as_batch_sampler=True``.
            as_batch_sampler (bool, optional): If ``True``, this instance
                will behave like :cls:`torch.utils.data.BatchSampler.
                This is highly recommended if ``batched_by`` if specified.
            drop_last (bool, optional): If ``True``, drop the last batch for each group.
                If ``batch_by`` is specified, this will also drop the last batch for each group
                if it does not meet the specifications.
            shuffle (bool, optional): If ``True``, shuffles the data.
            seed (torch.Tensor): Random seed to use for initialization.
        """
        group_by = None

        super().__init__(dataset)
        logger = logging.getLogger(__name__)

        if batch_by and not batch_size:
            raise ValueError("`batch_size` must be specified if `batched=True`")
        if batch_by and not as_batch_sampler:
            logger.warning(
                "Using `batch_by` without batch sampling functionality. "
                "To use as a batch sampler, set `as_batch_sampler=True`."
            )

        # Configure batch by groups if they exist.
        group_unknown = False
        if batch_by:
            groups = _build_groups(dataset.examples, batch_by, group_unknown=group_unknown)
            if group_by:
                groups = {
                    grp: _build_groups(
                        dataset.examples, group_by, group_unknown=group_unknown, indices=idxs
                    )
                    for grp, idxs in groups.items()
                }
            else:
                groups = {k: {_UNKNOWN_TOKEN: v} for k, v in groups.items()}
        else:
            groups = {
                _UNKNOWN_TOKEN: _build_groups(
                    dataset.examples, group_by, group_unknown=group_unknown
                )
            }

        if seed is None:
            rng = torch.Generator().set_state(torch.random.get_rng_state())
        else:
            rng = torch.Generator().manual_seed(seed)

        self.group_by = group_by
        self.batch_by = batch_by
        self.batch_size = batch_size
        self.as_batch_sampler = as_batch_sampler
        self.drop_last = drop_last
        self.shuffle = shuffle

        self._rng = rng
        self._groups = groups
        batches = self._build_batches(shuffle=False)
        self._length = (
            len(batches) if self.as_batch_sampler else len(torch.cat(batches, dim=0).tolist())
        )

    def __len__(self):
        return self._length

    def __iter__(self):
        batches: List[torch.Tensor] = self._build_batches()

        if self.as_batch_sampler:
            return iter(batches)
        else:
            return iter(torch.cat(batches, dim=0).tolist())

    def _build_batches(self, shuffle: bool = None):
        """Build batches of examples to sample."""
        # Shuffle indices (if applicable).
        shuffle = self.shuffle if shuffle is None else shuffle
        if shuffle:
            groups = {k: _shuffle_groups(v, self._rng) for k, v in self._groups.items()}
        else:
            groups = self._groups

        # Break into chunks of <= batch_size.
        batch_size = self.batch_size if self.batch_size else 1
        if batch_size:
            groups = [
                torch.split(torch.as_tensor(v), batch_size)
                for grp in groups.values()
                for v in grp.values()
            ]
            # Drop last
            if self.drop_last:
                groups = [v[:-1] if len(v[-1]) < batch_size else v for v in groups]

        batches = [batch for grp in groups for batch in grp]

        # Reorder groups if shuffle is enabled.
        if shuffle:
            batches = [batches[i] for i in torch.randperm(len(batches), generator=self._rng)]

        return batches


class AlternatingGroupSampler(GroupSampler):
    """Alternate sampler between supervised/unsupervised examples within groups."""

    def __init__(
        self,
        dataset,
        T_s: int = 1,
        T_us: int = 1,
        batch_by: Any = None,
        batch_size: int = None,
        as_batch_sampler: bool = False,
        drop_last: bool = False,
        seed: int = None,
    ):
        self.T_s = T_s
        self.T_us = T_us

        example = dataset.examples[0]
        key = [k for k in ("is_unsupervised", "_is_unsupervised") if k in example]
        assert len(key) == 1
        key = key[0]

        if batch_by is None:
            batch_by = ()
        batch_by = tuple(batch_by) if isinstance(batch_by, (list, tuple)) else (batch_by,)
        batch_by += (key,)

        super().__init__(
            dataset,
            batch_by=batch_by,
            batch_size=batch_size if batch_size else 1,
            as_batch_sampler=as_batch_sampler,
            drop_last=False,
            shuffle=True,
            seed=seed,
        )

        # Groups should only have one subgroup.
        counts = {}
        counts_per_group = defaultdict(int)
        for grp_name, grp in self._groups.items():
            if len(grp) > 1:
                raise ValueError(
                    f"All groups should only have 1 subgroup. "
                    f"{grp_name} has {len(grp)} - {grp.keys()}"
                )
            num = len(list(grp.values())[0])
            counts_per_group[grp_name[:-1]] += num
            counts[grp_name] = num
        # Drop any groups that will be smaller than the batch size.
        if drop_last:
            drop_keys = [grp_name for grp_name, count in counts.items() if count < batch_size]
            for k in drop_keys:
                counts_per_group.pop(k)
                for gk in [k + (True,), k + (False,)]:
                    counts.pop(gk, None)
                    self._groups.pop(gk, None)
        self.counts_per_group = counts_per_group
        self.counts = counts
        self._samplers = self._build_samplers(seed=seed)

        # State variables
        self._iterators = {}
        self._counter = {grp_name: 0 for grp_name in self._samplers}
        self._pointer = 0
        self._excess_sampled = 0

    def __len__(self):
        """
        Note:
            This property should not be used for any real logic.
            It is purely to satisfy the requirement that samplers
            used as BatchSamplers should have a length.
        """
        return sum(v // self.batch_size for v in self.counts.values())

    def _batch_by_groups(self):
        return list(self.counts_per_group.keys())

    def _build_samplers(self, seed=None) -> Dict[Tuple[Hashable], SubsetRandomSampler]:
        groups = self._groups
        samplers = {}
        for grp_idx, (grp_name, grp) in enumerate(groups.items()):
            indices = list(grp.values())[0]
            g_seed = seed + grp_idx if seed is not None else None
            gen = torch.Generator()
            if g_seed is not None:
                gen = gen.manual_seed(g_seed)
            samplers[grp_name] = SubsetRandomSampler(indices, generator=gen)
        return samplers

    def _sample(self, group: Hashable, unsupervised: bool):
        def _reset_iterator(grp):
            self._iterators[grp] = iter(self._samplers[grp])

        group = group + (bool(unsupervised),)

        if group not in self._iterators:
            _reset_iterator(group)

        try:
            idx = next(self._iterators[group])
        except StopIteration:
            _reset_iterator(group)
            idx = next(self._iterators[group])

        self._counter[group] += 1
        return idx

    def _get_group(self, excess_sample, to_sample: List[bool]):
        """Return the group that should be sampled for this batch.

        Args:
            excess_sample (int): A value ``<0`` indicates that a surplus of supervised
                examples have previously been sampled. The algorithm will then choose
                a group with probability based on the remaining number of unsupervised examples.
                If value is ``>0``, then vice versa.
        """
        batch_by = self._batch_by_groups()
        sample_by_groups = (
            [excess_sample < 0] if excess_sample != 0 else sorted(np.unique(to_sample))
        )

        # Find the relative epoch for each iterator.
        num_examples_by_group = np.asarray(
            [
                [self.counts.get(batch_grp + (x,), 0) for x in sample_by_groups]
                for batch_grp in batch_by
            ]
        )
        counter = np.asarray(
            [
                [self._counter.get(batch_grp + (x,), 0) for x in sample_by_groups]
                for batch_grp in batch_by
            ]
        )

        epoch_num = counter // (num_examples_by_group + (num_examples_by_group == 0))
        epoch_num[num_examples_by_group == 0] = np.max(epoch_num) + 1
        remainder = num_examples_by_group - counter % (
            num_examples_by_group + (num_examples_by_group == 0)
        )
        remainder[epoch_num != np.min(epoch_num)] = 0
        weights = np.sum(remainder, axis=1)
        p = torch.as_tensor(weights / np.sum(weights))

        idx = torch.multinomial(p, 1, generator=self._rng).item()
        return batch_by[idx]

    def _next_batch(self) -> List[int]:
        batch_size: int = self.batch_size
        # pointer must be in range [0, self.T_s + self.T_us)
        pointer = self._pointer
        # False = supervised, True = unsupervised
        to_sample: List[bool] = []

        num_excess_sample = min(abs(self._excess_sampled), batch_size)
        to_sample.extend([self._excess_sampled < 0] * num_excess_sample)

        num_alt_sample = batch_size - len(to_sample)
        c_range = np.arange(pointer, pointer + num_alt_sample)
        to_sample.extend(c_range % (self.T_s + self.T_us) >= self.T_s)

        # Find group to sample.
        group = self._get_group(self._excess_sampled, to_sample)

        # If sampled group only has supervised or unsupervised examples (but not both),
        # the batch will be composed of only supervised or only unsupervised examples.
        # This violates the duty cycle of supervised to unsupervised samples.
        # We change the self._excess_sampled state variable to indicate that a surplus
        # of either supervised or unsupervised samples has been selected. In this case,
        # the next batch will attempt to "correct" this sampling by oversampling the
        # opposite.
        is_unsupervised_options = [False, True]
        has_supervised_unsupervised = [
            group + (is_unsup,) in self._groups for is_unsup in is_unsupervised_options
        ]
        # TODO: Check if to_sample is made up of all of same type.
        if not all(has_supervised_unsupervised):
            idx = np.where(has_supervised_unsupervised)[0][0]
            is_unsup = is_unsupervised_options[idx]
            to_sample = [is_unsup] * batch_size
            shift_pointer = 0
            while is_unsup == ((pointer + shift_pointer) % (self.T_s + self.T_us) >= self.T_s):
                shift_pointer += 1
            self._excess_sampled += ((-1) ** (is_unsup + 1)) * (batch_size - shift_pointer)
            self._pointer += shift_pointer
        else:
            if self._excess_sampled < 0:
                self._excess_sampled = self._excess_sampled + num_excess_sample
            else:
                self._excess_sampled = self._excess_sampled - num_excess_sample
            self._pointer = (self._pointer + num_alt_sample) % (self.T_s + self.T_us)

        samples = torch.as_tensor([self._sample(group, is_unsup) for is_unsup in to_sample])
        return samples

    def _build_batches(self, shuffle: bool = None):
        """Build batches of examples to sample."""
        # Hacky way of getting around GroupSampler.__init__ _length computation.
        if not hasattr(self, "counts"):
            return [torch.tensor([])]

        return [self._next_batch() for _ in range(len(self))]


class DistributedGroupSampler(Sampler):
    """Samples examples such that examples from the same group are on the same node.

    `dataset` must support the following attributes and methods
        * `groups(group_by) -> Dict`: Returns mapping from group id to indices

    dataset examples is a list of tuples (fname, instance),
    where fname is essentially the volume name (actually a filename).
    """

    def __init__(self, dataset, group_by: str, shuffle: bool = False):
        self.dataset = dataset
        self.world_size = self._world_size()
        self.rank = self._rank()
        self.epoch = 0
        self._group_by = group_by
        self.shuffle = shuffle

        # All nodes
        all_groups = dataset.groups(group_by)
        self.all_groups = np.array(sorted(all_groups.keys()))
        self.all_groups_split = np.array_split(self.all_groups, self.world_size)

        # This node
        self.groups = self.all_groups_split[self.rank]
        self.indices = np.concatenate([all_groups[group_id] for group_id in self.groups])
        self.num_samples = len(self.indices)

    def _world_size(self):
        return comm.get_world_size()

    def _rank(self):
        return comm.get_rank()

    def __iter__(self):
        # deterministically shuffle based on epoch
        if self.shuffle:
            g = torch.Generator()
            g.manual_seed(self.epoch)
            ordering = torch.randperm(self.num_samples, generator=g).tolist()
            indices = self.indices[ordering]
        else:
            indices = self.indices
        return iter(indices.tolist())

    def __len__(self):
        return self.num_samples

    def set_epoch(self, epoch):
        self.epoch = epoch


def _shuffle_groups(groups, rng):
    if not isinstance(groups, Mapping):
        return torch.as_tensor(groups)[torch.randperm(len(groups), generator=rng)]

    return {k: _shuffle_groups(v, rng) for k, v in groups.items()}


def _build_groups(
    examples,
    group_by: Union[Any, Sequence[Any]] = None,
    group_unknown: bool = False,
    indices: Sequence[int] = None,
):
    if not indices:
        indices = range(len(examples))

    if group_by is None:
        return {_UNKNOWN_TOKEN: indices}
    elif isinstance(group_by, str):
        group_by = (group_by,)

    group_by_values = []
    for idx in indices:
        ex = examples[idx]
        values = []
        for gb in group_by:
            if gb in ex:
                values.append(ex[gb])
            elif gb in ex.get("_metadata", {}):
                values.append(ex["_metadata"][gb])
            else:
                values.append(_UNKNOWN_TOKEN)

        if group_unknown and _UNKNOWN_TOKEN in values:
            values = [_UNKNOWN_TOKEN]
        group_by_values.append(tuple(values))

    group_to_idxs = defaultdict(list)
    for idx, val in zip(indices, group_by_values):
        group_to_idxs[val].append(idx)

    return group_to_idxs



# flake8: noqa: F401

from meddlr.metrics import Metric
from meddlr.utils.cluster import Cluster

# This line will be programatically read/write by setup.py.
# Leave them at the bottom of this file and don't touch them.
__version__ = "0.0.1a5"



import inspect
import logging
from copy import deepcopy
from typing import Any, Dict

from meddlr.metrics.collection import MetricCollection
from meddlr.metrics.image import MSE, NRMSE, PSNR, RMSE, SSIM
from meddlr.metrics.sem_seg import ASSD, CV, DSC, VOE

logger = logging.getLogger(__name__)


# The builtin metrics that will be recognized by every
# evaluator based on the string completion.
# We specify the class and any keyword args required to
# construct the object.
# fmt: off
_BUILTIN_METRICS = {
    # =========== Image =========== #
    # PSNR
    "psnr": (PSNR, {"im_type": None}),
    "psnr_mag": (PSNR, {"im_type": "magnitude"}),
    "psnr_real": (PSNR, {"im_type": "real"}),
    "psnr_phase": (PSNR, {"im_type": "phase"}),
    "psnr_imag": (PSNR, {"im_type": "imag"}),
    # NRMSE
    "nrmse": (NRMSE, {"im_type": None}),
    "nrmse_mag": (NRMSE, {"im_type": "magnitude"}),
    "nrmse_real": (NRMSE, {"im_type": "real"}),
    "nrmse_phase": (NRMSE, {"im_type": "phase"}),
    "nrmse_imag": (NRMSE, {"im_type": "imag"}),
    # MSE
    "mse": (MSE, {"im_type": None}),
    "mse_mag": (MSE, {"im_type": "magnitude"}),
    "mse_real": (MSE, {"im_type": "real"}),
    "mse_phase": (MSE, {"im_type": "phase"}),
    "mse_imag": (MSE, {"im_type": "imag"}),
    # RMSE
    "rmse": (RMSE, {"im_type": None}),
    "rmse_mag": (RMSE, {"im_type": "magnitude"}),
    "rmse_real": (RMSE, {"im_type": "real"}),
    "rmse_phase": (RMSE, {"im_type": "phase"}),
    "rmse_imag": (RMSE, {"im_type": "imag"}),
    # SSIM
    "ssim (Wang)": (SSIM, {"method": "wang"}),

    # =========== Semantic Segmentation =========== #
    "DSC": (DSC, {}),
    "VOE": (VOE, {}),
    "ASSD": (ASSD, {"connectivity": 1}),
    "CV": (CV, {})
}
# fmt: on


def build_metrics(metric_names, fmt: str = None, **kwargs) -> MetricCollection:
    metrics = {}
    for name in metric_names:
        if name not in _BUILTIN_METRICS.keys():
            raise ValueError(
                "Metric '{}' not found in built-in metrics. "
                "Built-in metrics include:\n{}".format(name, _BUILTIN_METRICS.keys())
            )

        klass: type
        base_kwargs: Dict[str, Any]
        klass, base_kwargs = _BUILTIN_METRICS[name]
        signature = inspect.signature(klass)
        klass_kwargs = {k: v for k, v in kwargs.items() if k in signature.parameters}

        shared_keys = base_kwargs.keys() & klass_kwargs.keys()
        if shared_keys:
            logger.warning(
                "Metric '{}' default values for following arguments will be overridden. "
                "This is may lead to non-reproducible results:\n{}".format(name, shared_keys)
            )
        base_kwargs = deepcopy(base_kwargs)
        base_kwargs.update(klass_kwargs)

        mname = fmt.format(name) if fmt else name
        metrics[mname] = klass(**base_kwargs)

    return MetricCollection(metrics)



from .train_loop import *  # noqa

__all__ = [k for k in globals().keys() if not k.startswith("_")]


from .defaults import *  # noqa

# prefer to let hooks and defaults live in separate namespaces (therefore not in __all__)
# but still make them available here
from .hooks import *  # noqa
from .trainer import *  # noqa



from typing import Sequence

import torch

import meddlr.metrics.functional as mF
from meddlr.metrics.metric import Metric

__all__ = ["DSC", "CV", "VOE", "ASSD"]


class DSC(Metric):
    """Dice score coefficient.

    Attributes:
        channel_names (Sequence[str]): Category names corresponding to the channels.
    """

    is_differentiable = True
    higher_is_better = True

    def __init__(
        self,
        channel_names: Sequence[str] = None,
        reduction="none",
        compute_on_step: bool = False,
        dist_sync_on_step: bool = False,
        process_group: bool = None,
        dist_sync_fn: bool = None,
    ):
        super().__init__(
            channel_names=channel_names,
            units="",
            reduction=reduction,
            compute_on_step=compute_on_step,
            dist_sync_on_step=dist_sync_on_step,
            process_group=process_group,
            dist_sync_fn=dist_sync_fn,
        )

    def func(self, preds, targets) -> torch.Tensor:
        return mF.dice_score(y_pred=preds, y_true=targets)


Dice = DSC


class CV(Metric):
    """Coefficient of variation.

    Attributes:
        channel_names (Sequence[str]): Category names corresponding to the channels.
    """

    is_differentiable = True
    higher_is_better = False

    def __init__(
        self,
        channel_names: Sequence[str] = None,
        reduction="none",
        compute_on_step: bool = False,
        dist_sync_on_step: bool = False,
        process_group: bool = None,
        dist_sync_fn: bool = None,
    ):
        super().__init__(
            channel_names=channel_names,
            units="",
            reduction=reduction,
            compute_on_step=compute_on_step,
            dist_sync_on_step=dist_sync_on_step,
            process_group=process_group,
            dist_sync_fn=dist_sync_fn,
        )

    def func(self, preds, targets) -> torch.Tensor:
        return mF.coefficient_variation(y_pred=preds, y_true=targets)


class VOE(Metric):
    """Volumetric overlap error.

    Attributes:
        channel_names (Sequence[str]): Category names corresponding to the channels.
    """

    is_differentiable = True
    higher_is_better = False

    def __init__(
        self,
        channel_names: Sequence[str] = None,
        reduction="none",
        compute_on_step: bool = False,
        dist_sync_on_step: bool = False,
        process_group: bool = None,
        dist_sync_fn: bool = None,
    ):
        super().__init__(
            channel_names=channel_names,
            units="",
            reduction=reduction,
            compute_on_step=compute_on_step,
            dist_sync_on_step=dist_sync_on_step,
            process_group=process_group,
            dist_sync_fn=dist_sync_fn,
        )

    def func(self, preds, targets) -> torch.Tensor:
        return mF.volumetric_overlap_error(y_pred=preds, y_true=targets)


class ASSD(Metric):
    """Average symmetric surface distance.

    Attributes:
        connectivity (int): The neighbourhood/connectivity considered when determining
            the surface of the binary objects.
        channel_names (Sequence[str]): Category names corresponding to the channels.

    Note:
        This metric is not differentiable.
    """

    is_differentiable = False
    higher_is_better = False

    def __init__(
        self,
        connectivity: int = 1,
        channel_names: Sequence[str] = None,
        reduction="none",
        compute_on_step: bool = False,
        dist_sync_on_step: bool = False,
        process_group: bool = None,
        dist_sync_fn: bool = None,
    ):
        """
        Args:
            connectivity (int): The neighbourhood/connectivity considered when determining
                the surface of the binary objects. If in doubt, leave it as it is.
            channel_names (Sequence[str]): Category names corresponding to the channels.
        """
        super().__init__(
            channel_names=channel_names,
            units="",
            reduction=reduction,
            compute_on_step=compute_on_step,
            dist_sync_on_step=dist_sync_on_step,
            process_group=process_group,
            dist_sync_fn=dist_sync_fn,
        )
        self.connectivity = connectivity

    def func(self, preds, targets, spacing=None) -> torch.Tensor:
        return mF.assd(
            y_pred=preds, y_true=targets, spacing=spacing, connectivity=self.connectivity
        )

    def update(self, preds, targets, spacing=None, ids=None):
        return super().update(preds, targets, spacing=spacing, ids=ids)



from typing import Any, Dict, List, Optional, Sequence, Set, Union

import numpy as np
import pandas as pd
import tabulate
from torchmetrics.collections import MetricCollection as _MetricCollection

from meddlr.metrics.metric import Metric

__all__ = ["MetricCollection"]


class MetricCollection(_MetricCollection):
    """The class that manages multiple metrics."""

    def __init__(
        self,
        metrics: Union[Metric, Sequence[Metric], Dict[str, Metric]],
        *additional_metrics: Metric,
        prefix: Optional[str] = None,
        postfix: Optional[str] = None
    ) -> None:
        super().__init__(metrics, *additional_metrics, prefix=prefix, postfix=postfix)
        self._is_data_stale = False

    def scans(self) -> List[str]:
        return list(self._scan_data.keys())

    def scan_summary(self, scan_id, delimiter: str = ", ") -> str:
        """Get summary of results for a scan.
        Args:
            scan_id: Scan id for which to summarize results.
            delimiter (`str`, optional): Delimiter between different metrics.
        Returns:
            str: A summary of metrics for the scan. Values are averaged across
                all categories.
        """
        scan_data = self._scan_data[scan_id]
        avg_data = scan_data.mean(axis=1)

        strs = ["{}: {:0.3f}".format(n, avg_data[n]) for n in avg_data.index.tolist()]

        return delimiter.join(strs)

    def to_pandas(self, sync_dist: bool = True) -> pd.DataFrame:
        frames = []
        metric: Metric
        for name, metric in self.items():
            df: pd.DataFrame = metric.to_pandas(sync_dist=sync_dist)
            df["Metric"] = name
            frames.append(df)

        return pd.concat(frames, ignore_index=True)

    def to_dict(self, group_by="Metric", sync_dist: bool = True) -> Dict[str, Any]:
        df = self.to_pandas(sync_dist=sync_dist)
        df = df.melt(id_vars=["Metric", "id"], var_name="category", value_name="value")
        if len(np.unique(df["category"])) > 1:
            df["Metric"] = df["Metric"] + "/" + df["category"]
        df = df.drop(columns="category")
        values = df.groupby(by=group_by).mean()
        return values.to_dict()["value"]

    def summary(self, sync_dist: bool = True) -> str:
        """Get summary of results overall scans.
        Returns:
            str: Tabulated summary. Rows=metrics. Columns=classes.
        """
        df = self.to_pandas(sync_dist=sync_dist)
        if "id" in df:
            df = df.drop(columns="id")
        df = df.groupby(by="Metric")

        mean = df.mean().applymap(lambda x: "{:0.3f}".format(x))
        std = df.std().applymap(lambda x: "{:0.3f}".format(x))
        df = mean + " (" + std + ")"
        return tabulate.tabulate(df, headers=df.columns) + "\n"

    def ids(self, sync_dist=True) -> Set[str]:
        _ids = set()
        metric: Metric
        for _, metric in self.items():
            _ids |= set(metric.to_pandas(sync_dist=sync_dist)["id"].to_numpy())
        return _ids



import logging
import math
import os
import re
from collections import OrderedDict
from typing import Mapping, Sequence, Union

from torch.nn import DataParallel

from meddlr.checkpoint import Checkpointer
from meddlr.config import CfgNode
from meddlr.data import build_recon_train_loader, build_recon_val_loader
from meddlr.engine import SimpleTrainer, hooks
from meddlr.evaluation import (
    DatasetEvaluator,
    ReconEvaluator,
    inference_on_dataset,
    print_csv_format,
    verify_results,
)
from meddlr.modeling import build_loss_computer, build_model
from meddlr.solver import GradAccumOptimizer, build_lr_scheduler, build_optimizer
from meddlr.utils import env
from meddlr.utils.events import CommonMetricPrinter, JSONWriter, TensorboardXWriter
from meddlr.utils.logger import setup_logger

__all__ = ["DefaultTrainer"]


def format_as_iter(vals: Union[int, Sequence[int]], iters_per_epoch: int, time_scale: str):
    """Format data to be at iteration time scale.

    If values are negative, they correspond to the opposite time scale.
    For example if `time_scale="epoch"` and `val=-1`, `val` corresponds to
    1 iteration.

    Args:
        vals (`int(s)`): Values to format.
        iters_per_epoch (int): Number of iterations per epoch.
        time_scale (str): Time scale of current values.

    Returns:
        int: Time (positive int) formatted in iteration format
    """
    assert time_scale in ["epoch", "iter"]

    single_value = not isinstance(vals, Sequence)
    if single_value:
        vals = [vals]

    epoch_convention = (time_scale == "epoch" and all(x >= 0 for x in vals)) or (
        time_scale == "iter" and all(x <= 0 for x in vals)
    )

    if epoch_convention:
        vals = type(vals)([iters_per_epoch * abs(x) for x in vals])
    else:
        vals = type(vals)([abs(x) for x in vals])

    if single_value:
        return vals[0]
    else:
        return vals


def _convert_time_recursive(
    entity: Union[Mapping, Sequence],
    iters_per_epoch,
    time_scale,
    patterns=("^.*_iters$", "^.*_iter$", "^.*_milestones$"),
):
    otype = type(entity)
    if isinstance(entity, (list, tuple)):
        out = []
        for e in entity:
            out.append(_convert_time_recursive(e, iters_per_epoch, time_scale, patterns))
        return otype(out)
    elif isinstance(entity, Mapping):
        entity = {k: v for k, v in entity.items()}
        for k in entity.keys():
            if not isinstance(entity[k], Mapping):
                if any(re.match(pattern, k) for pattern in patterns):
                    entity[k] = format_as_iter(entity[k], iters_per_epoch, time_scale)
                elif not isinstance(entity[k], (list, tuple)):
                    continue

            entity[k] = _convert_time_recursive(entity[k], iters_per_epoch, time_scale, patterns)
        return otype(entity)
    else:
        return entity


def convert_cfg_time_to_iter(cfg: CfgNode, iters_per_epoch: int, ignore_missing: bool = False):
    """Convert all config time-related parameters to iteration scale.

    Args:
        cfg (CfgNode): The config to
        iters_per_epoch (int): Number of iterations per epoch.
        ignore_missing (bool, optional): If ``True``, silently skip over missing keys.
    """
    cfg = cfg.clone()
    cfg.defrost()

    time_scale = cfg.TIME_SCALE

    time_fmt = (
        "SOLVER.MAX_ITER",
        "SOLVER.STEPS",
        "SOLVER.CHECKPOINT_PERIOD",
        "TEST.EVAL_PERIOD",
        "VIS_PERIOD",
        "MODEL.CONSISTENCY.AUG.NOISE.SCHEDULER.WARMUP_ITERS",
    )
    recursively_fmt = (
        "AUG_TRAIN.MRI_RECON.TRANSFORMS",
        "AUG_TRAIN.MRI_RECON.SCHEDULER_P",
        "MODEL.CONSISTENCY.AUG.MRI_RECON.TRANSFORMS",
        "MODEL.CONSISTENCY.AUG.MRI_RECON.SCHEDULER_P",
    )

    for keys_to_fmt, func in [
        (time_fmt, format_as_iter),
        (recursively_fmt, _convert_time_recursive),
    ]:
        for key in keys_to_fmt:
            try:
                curr_val = cfg.get_recursive(key)
            except KeyError:
                if ignore_missing:
                    continue
                raise KeyError(f"cfg does not contain key '{key}'")
            value = func(curr_val, iters_per_epoch, time_scale)
            cfg.set_recursive(key, value)

    cfg.TIME_SCALE = "iter"
    cfg.freeze()
    return cfg


class DefaultTrainer(SimpleTrainer):
    """
    A trainer with default training logic. Compared to `SimpleTrainer`, it
    contains the following logic in addition:

    1. Create model, optimizer, scheduler, dataloader from the given config.
    2. Load a checkpoint or `cfg.MODEL.WEIGHTS`, if exists.
    3. Register a few common hooks.

    It is created to simplify the **standard model training workflow**
    and reduce code boilerplate for users who only need the standard training
    workflow, with standard features.
    It means this class makes *many assumptions* about your training logic that
    may easily become invalid in a new research. In fact, any assumptions beyond
    those made in the :class:`SimpleTrainer` are too much for research.

    The code of this class has been annotated about restrictive assumptions it
    makes. When they do not work for you, you're encouraged to:

    1. Overwrite methods of this class, OR:
    2. Use :class:`SimpleTrainer`, which only does minimal SGD training and
       nothing else. You can then add your own hooks if needed. OR:
    3. Write your own training loop similar to `tools/plain_train_net.py`.

    Also note that the behavior of this class, like other functions/classes in
    this file, is not stable, since it is meant to represent the
    "common default behavior".
    It is only guaranteed to work well with the standard models and training
    workflow in meddlr.
    To obtain more stable behavior, write your own training logic with other
    public APIs.

    Attributes:
        scheduler:
        checkpointer (DetectionCheckpointer):
        cfg (CfgNode):

    Examples:

    .. code-block:: python

        trainer = DefaultTrainer(cfg)
        trainer.resume_or_load()  # load last checkpoint or MODEL.WEIGHTS
        trainer.train()
    """

    def __init__(self, cfg):
        """
        Args:
            cfg (CfgNode):
        """
        logger = logging.getLogger("meddlr")
        if not logger.isEnabledFor(logging.INFO):  # setup_logger is not called for d2
            setup_logger()

        # Assume these objects must be constructed in this order.
        data_loader = self.build_train_loader(cfg)
        num_iter_per_epoch = len(data_loader.dataset) / cfg.SOLVER.TRAIN_BATCH_SIZE
        if cfg.DATALOADER.DROP_LAST:
            num_iter_per_epoch = int(num_iter_per_epoch)
        else:
            num_iter_per_epoch = math.ceil(num_iter_per_epoch)
        cfg = convert_cfg_time_to_iter(cfg, num_iter_per_epoch)
        logger.info(f"Calculated {num_iter_per_epoch} iterations per epoch")

        # Recreate data loader in case data loader required epoch -> iter conversion.
        del data_loader
        data_loader = self.build_train_loader(cfg)

        model = self.build_model(cfg)
        optimizer = self.build_optimizer(cfg, model)

        # For training, wrap with DP. But don't need this for inference.
        num_gpus = len(os.environ["CUDA_VISIBLE_DEVICES"].split(","))
        if num_gpus > 1:
            logger.info("Using data parallel...")
            model = DataParallel(model)
        model.to(cfg.MODEL.DEVICE)

        loss_computer = self.build_loss_computer(cfg)
        super().__init__(model, data_loader, optimizer, loss_computer)

        self.scheduler = self.build_lr_scheduler(cfg, optimizer)
        # Assume no other objects need to be checkpointed.
        # We can later make it checkpoint the stateful hooks
        self.checkpointer = Checkpointer(
            # Assume you want to save checkpoints together with logs/statistics
            model,
            cfg.OUTPUT_DIR,
            optimizer=optimizer,
            scheduler=self.scheduler,
        )
        self.start_iter = 0
        self.max_iter = cfg.SOLVER.MAX_ITER
        self.cfg = cfg

        self.register_hooks(self.build_hooks())

    def init_fine_tune(self):
        """
        Initialize model with fine-tune weights.

        If fine tune weights not specified, results in no-op.
        """
        fine_tune_weights = self.cfg.MODEL.FINE_TUNE.WEIGHTS
        if not fine_tune_weights:
            return

        logger = logging.getLogger(__name__)
        logger.info("Loading fine-tune weights")
        if not os.path.isfile(fine_tune_weights):
            raise FileNotFoundError("weights not found: {}".format(fine_tune_weights))

        temp_checkpointer = Checkpointer(self.model)
        temp_checkpointer.load(fine_tune_weights)

    def resume_or_load(self, resume=True, restart_iter=False):
        """
        If `resume==True`, and last checkpoint exists, resume from it.

        Otherwise, load a model specified by the config.

        Args:
            resume (bool): whether to do resume or not
        """
        # The checkpoint stores the training iteration that just finished,
        # thus we start at the next iteration
        # (or iter zero if there's no checkpoint).
        iteration = (
            self.checkpointer.resume_or_load(self.cfg.MODEL.WEIGHTS, resume=resume).get(
                "iteration", -1
            )
            + 1
        )
        self.start_iter = iteration if not restart_iter else 0

    def build_hooks(self):
        """
        Build a list of default hooks, including timing, evaluation,
        checkpointing, lr scheduling, precise BN, writing events.

        Returns:
            list[HookBase]:
        """
        cfg = self.cfg.clone()
        cfg.defrost()
        cfg.DATALOADER.NUM_WORKERS = 0  # save some memory and time for PreciseBN

        ret = [
            (
                hooks.FlushOptimizer(self.optimizer, cfg.TEST.EVAL_PERIOD)
                if isinstance(self.optimizer, GradAccumOptimizer)
                else None
            ),
            hooks.IterationTimer(),
            hooks.LRScheduler(self.optimizer, self.scheduler),
            hooks.PeriodicCheckpointer(self.checkpointer, cfg.SOLVER.CHECKPOINT_PERIOD),
        ]

        def test_and_save_results():
            self._last_eval_results = self.test(self.cfg, self.model, use_val=True)
            return self._last_eval_results

        # Do evaluation after checkpointer, because then if it fails,
        # we can use the saved checkpoint to debug.
        ret.append(hooks.EvalHook(cfg.TEST.EVAL_PERIOD, test_and_save_results))
        if env.profile_memory():
            ret.append(hooks.MemoryProfiler(20))

        # Writing should be last
        ret.append(hooks.PeriodicWriter(self.build_writers(), periods=(20, cfg.TEST.EVAL_PERIOD)))
        ret = [x for x in ret if x is not None]
        return ret

    def build_writers(self):
        """
        Build a list of writers to be used. By default it contains
        writers that write metrics to the screen,
        a json file, and a tensorboard event file respectively.
        If you'd like a different list of writers, you can overwrite it in
        your trainer.

        Returns:
            list[EventWriter]: a list of :class:`EventWriter` objects.

        It is now implemented by:

        .. code-block:: python

            return [
                CommonMetricPrinter(self.max_iter),
                JSONWriter(os.path.join(self.cfg.OUTPUT_DIR, "metrics.json")),
                TensorboardXWriter(self.cfg.OUTPUT_DIR),
            ]

        """
        assert (
            self.cfg.TIME_SCALE == "iter" and self.cfg.TEST.EVAL_PERIOD >= 0
        ), "cfg should be formatted in 'iter' time scale"
        eval_period = self.cfg.TEST.EVAL_PERIOD

        # Assume the default print/log frequency.
        return [
            # It may not always print what you want to see,
            # since it prints "common" metrics only.
            CommonMetricPrinter(self.max_iter, eval_period=eval_period),
            JSONWriter(os.path.join(self.cfg.OUTPUT_DIR, "metrics.json")),
            TensorboardXWriter(self.cfg.OUTPUT_DIR),
        ]

    def train(self):
        """
        Run training.

        Returns:
            OrderedDict of results, if evaluation is enabled. Otherwise None.
        """
        super().train(self.start_iter, self.max_iter)
        if hasattr(self, "_last_eval_results"):
            verify_results(self.cfg, self._last_eval_results)
            return self._last_eval_results

    @classmethod
    def build_model(cls, cfg):
        """
        Returns:
            torch.nn.Module:

        It now calls :func:`meddlr.modeling.build_model`.
        Overwrite it if you'd like a different model.
        """
        model = build_model(cfg)
        # logger = logging.getLogger(__name__)
        # printing model doesnt really help
        # logger.info("Model:\n{}".format(model))
        return model

    @classmethod
    def build_loss_computer(cls, cfg):
        loss_computer = (
            "N2RLossComputer"
            if cfg.MODEL.META_ARCHITECTURE == "N2RModel"
            or cfg.MODEL.META_ARCHITECTURE == "M2RModel"
            or cfg.MODEL.META_ARCHITECTURE == "NM2RModel"
            or cfg.MODEL.META_ARCHITECTURE == "A2RModel"
            else "BasicLossComputer"
        )
        return build_loss_computer(cfg, loss_computer)

    @classmethod
    def build_optimizer(cls, cfg, model):
        """
        Returns:
            torch.optim.Optimizer:

        It now calls :func:`meddlr.solver.build_optimizer`.
        Overwrite it if you'd like a different optimizer.
        """
        return build_optimizer(cfg, model)

    @classmethod
    def build_lr_scheduler(cls, cfg, optimizer):
        """
        It now calls :func:`meddlr.solver.build_lr_scheduler`.
        Overwrite it if you'd like a different scheduler.
        """
        return build_lr_scheduler(cfg, optimizer)

    @classmethod
    def build_train_loader(cls, cfg):
        """
        Returns:
            iterable

        It now calls :func:`meddlr.data.build_recon_train_loader`.
        Overwrite it if you'd like a different data loader.
        """
        return build_recon_train_loader(cfg)

    @classmethod
    def build_test_loader(cls, cfg, dataset_name, is_val=False):
        """
        Returns:
            iterable

        It now calls :func:`meddlr.data.build_recon_val_loader`.
        Overwrite it if you'd like a different data loader.
        """
        if is_val:
            as_test = cfg.TEST.VAL_AS_TEST
        else:
            as_test = True
        return build_recon_val_loader(cfg, dataset_name, as_test=as_test)

    @classmethod
    def build_evaluator(cls, cfg, dataset_name, is_val=False):
        """
        Returns:
            DatasetEvaluator

        It is not implemented by default.
        """
        if is_val and len(cfg.TEST.VAL_METRICS.RECON) > 0:
            metrics = cfg.TEST.VAL_METRICS.RECON
        else:
            metrics = None
        return ReconEvaluator(dataset_name, cfg, metrics=metrics)

    @classmethod
    def test(cls, cfg, model, evaluators=None, use_val: bool = False):
        """
        Args:
            cfg (CfgNode):
            model (nn.Module):
            evaluators (list[DatasetEvaluator] or None): if None, will call
                :meth:`build_evaluator`. Otherwise, must have the same length
                as `cfg.DATASETS.{VAL/TEST}` depending on `use_val` flag.
            use_val (bool, optional): If `True`, un inference on validation data.
                Should be `True` during training.

        Returns:
            dict: a dict of result metrics
        """
        logger = logging.getLogger(__name__)
        if isinstance(evaluators, DatasetEvaluator):
            evaluators = [evaluators]

        # To use validation data, flag must be enabled and val datasets must be available.
        if use_val and cfg.DATASETS.VAL:
            inf_dataset = cfg.DATASETS.VAL
        else:
            inf_dataset = cfg.DATASETS.TEST
        if evaluators is not None:
            assert len(inf_dataset) == len(evaluators), "{} != {}".format(
                len(inf_dataset), len(evaluators)
            )

        results = OrderedDict()
        for idx, dataset_name in enumerate(inf_dataset):
            data_loader = cls.build_test_loader(cfg, dataset_name)
            # When evaluators are passed in as arguments,
            # implicitly assume that evaluators can be created
            # before data_loader.
            if evaluators is not None:
                evaluator = evaluators[idx]
            else:
                try:
                    evaluator = cls.build_evaluator(cfg, dataset_name, is_val=use_val)
                except NotImplementedError:
                    logger.warning(
                        "No evaluator found. "
                        "Use `DefaultTrainer.test(evaluators=)`, "
                        "or implement its `build_evaluator` method."
                    )
                    results[dataset_name] = {}
                    continue
            results_i = inference_on_dataset(model, data_loader, evaluator)
            results[dataset_name] = results_i
            assert isinstance(results_i, dict), (
                "Evaluator must return a dict or list on the main process. "
                "Got {} instead.".format(results_i)
            )
            logger.info("Evaluation results for {} in csv format:".format(dataset_name))
            print_csv_format(results_i)

        # if len(results) == 1:
        #     results = list(results.values())[0]
        return results



from meddlr.metrics import collection, image, metric, sem_seg
from meddlr.metrics.collection import MetricCollection  # noqa: F401
from meddlr.metrics.image import MSE, NRMSE, PSNR, RMSE, SSIM  # noqa: F401
from meddlr.metrics.metric import Metric  # noqa: F401
from meddlr.metrics.sem_seg import ASSD, CV, DSC, VOE  # noqa: F401

__all__ = []
__all__.extend(collection.__all__)
__all__.extend(metric.__all__)
__all__.extend(image.__all__)
__all__.extend(sem_seg.__all__)



from numbers import Number
from typing import Sequence, Union

import torch
import torchvision.utils as tv_utils
from torch import nn

import meddlr.ops.complex as cplx
from meddlr.config.config import configurable
from meddlr.utils.events import get_event_storage
from meddlr.utils.general import move_to_device
from meddlr.utils.transforms import SenseModel

from ..layers.layers2D import ResNet
from .build import META_ARCH_REGISTRY, build_model

__all__ = ["GeneralizedUnrolledCNN"]


@META_ARCH_REGISTRY.register()
class GeneralizedUnrolledCNN(nn.Module):
    """PyTorch implementation of Unrolled Compressed Sensing.

    This file contains an implementation of the Unrolled Compressed Sensing
    framework by CM Sandino, JY Cheng, et al. See paper below for more details.

    It is also based heavily on the codebase below:
    https://github.com/MRSRL/dl-cs

    Implementation is based on:
        CM Sandino, JY Cheng, et al. "Compressed Sensing: From Research to
        Clinical Practice with Deep Neural Networks" IEEE Signal Processing
        Magazine, 2020.
    """

    @configurable
    def __init__(
        self,
        blocks: nn.ModuleList,
        step_sizes: Union[float, Sequence[float]] = -2.0,
        fix_step_size: bool = False,
        num_emaps: int = 1,
        vis_period: int = -1,
    ):
        super().__init__()

        if isinstance(blocks, Sequence) and not isinstance(blocks, nn.ModuleList):
            blocks = nn.ModuleList(blocks)
        if not isinstance(blocks, nn.ModuleList):
            raise TypeError("`blocks` must be a sequence of nn.Modules or a nn.ModuleList")
        self.resnets = blocks
        num_grad_steps = len(blocks)

        if isinstance(step_sizes, Number):
            step_sizes = [
                torch.tensor([step_sizes], dtype=torch.float32) for _ in range(num_grad_steps)
            ]
        else:
            step_sizes = [torch.tensor(s) for s in step_sizes]
        if not fix_step_size:
            step_sizes = nn.ParameterList([nn.Parameter(s) for s in step_sizes])
        self.step_sizes: Sequence[Union[torch.Tensor, nn.Parameter]] = step_sizes

        self.num_emaps = num_emaps
        self.vis_period = vis_period

    def visualize_training(self, kspace, zfs, targets, preds):
        """A function used to visualize reconstructions.

        Args:
            targets: NxHxWx2 tensors of target images.
            preds: NxHxWx2 tensors of predictions.
        """
        storage = get_event_storage()

        with torch.no_grad():
            if cplx.is_complex(kspace):
                kspace = torch.view_as_real(kspace)
            kspace = kspace[0, ..., 0, :].unsqueeze(0).cpu()  # calc mask for first coil only
            targets = targets[0, ...].unsqueeze(0).cpu()
            preds = preds[0, ...].unsqueeze(0).cpu()
            zfs = zfs[0, ...].unsqueeze(0).cpu()

            all_images = torch.cat([zfs, preds, targets], dim=2)

            imgs_to_write = {
                "phases": cplx.angle(all_images),
                "images": cplx.abs(all_images),
                "errors": cplx.abs(preds - targets),
                "masks": cplx.get_mask(kspace),
            }

            for name, data in imgs_to_write.items():
                data = data.squeeze(-1).unsqueeze(1)
                data = tv_utils.make_grid(data, nrow=1, padding=1, normalize=True, scale_each=True)
                storage.put_image("train/{}".format(name), data.numpy(), data_format="CHW")

    def forward(self, inputs, return_pp=False, vis_training=False):
        """
        TODO: condense into list of dataset dicts.
        Args:
            inputs: Standard ss_recon module input dictionary
                * "kspace": Kspace. If fully sampled, and want to simulate
                    undersampled kspace, provide "mask" argument.
                * "maps": Sensitivity maps
                * "target" (optional): Target image (typically fully sampled).
                * "mask" (optional): Undersampling mask to apply.
                * "signal_model" (optional): The signal model. If provided,
                    "maps" will not be used to estimate the signal model.
                    Use with caution.
            return_pp (bool, optional): If `True`, return post-processing
                parameters "mean", "std", and "norm" if included in the input.
            vis_training (bool, optional): If `True`, force visualize training
                on this pass. Can only be `True` if model is in training mode.

        Returns:
            Dict: A standard ss_recon output dict
                * "pred": The reconstructed image
                * "target" (optional): The target image.
                    Added if provided in the input.
                * "mean"/"std"/"norm" (optional): Pre-processing parameters.
                    Added if provided in the input.
                * "zf_image": The zero-filled image.
                    Added when model is in eval mode.
        """
        if vis_training and not self.training:
            raise ValueError("vis_training is only applicable in training mode.")
        # Need to fetch device at runtime for proper data transfer.
        device = next(self.resnets[0].parameters()).device
        inputs = move_to_device(inputs, device)
        kspace = inputs["kspace"]
        target = inputs.get("target", None)
        mask = inputs.get("mask", None)
        A = inputs.get("signal_model", None)
        maps = inputs["maps"]
        num_maps_dim = -2 if cplx.is_complex_as_real(maps) else -1
        if self.num_emaps != maps.size()[num_maps_dim]:
            raise ValueError("Incorrect number of ESPIRiT maps! Re-prep data...")

        # Move step sizes to the right device.
        step_sizes = [x.to(device) for x in self.step_sizes]

        if mask is None:
            mask = cplx.get_mask(kspace)
        kspace *= mask

        # Get data dimensions
        dims = tuple(kspace.size())

        # Declare signal model.
        if A is None:
            A = SenseModel(maps, weights=mask)

        # Compute zero-filled image reconstruction
        zf_image = A(kspace, adjoint=True)

        # Begin unrolled proximal gradient descent
        image = zf_image
        for resnet, step_size in zip(self.resnets, step_sizes):
            # dc update
            grad_x = A(A(image), adjoint=True) - zf_image
            image = image + step_size * grad_x

            # If the image is a complex tensor, we view it as a real image
            # where last dimension has 2 channels (real, imaginary).
            # This may take more time, but is done for backwards compatibility
            # reasons.
            # TODO (arjundd): Fix to auto-detect which version of the model is
            # being used.
            use_cplx = cplx.is_complex(image)
            if use_cplx:
                image = torch.view_as_real(image)

            # prox update
            image = image.reshape(dims[0:3] + (self.num_emaps * 2,)).permute(0, 3, 1, 2)
            if hasattr(resnet, "base_forward") and callable(resnet.base_forward):
                image = resnet.base_forward(image)
            else:
                image = resnet(image)

            image = image.permute(0, 2, 3, 1).reshape(dims[0:3] + (self.num_emaps, 2))
            if not image.is_contiguous():
                image = image.contiguous()
            if use_cplx:
                image = torch.view_as_complex(image)

        output_dict = {"pred": image, "target": target}  # N x Y x Z x 1 x 2  # N x Y x Z x 1 x 2

        if return_pp:
            output_dict.update({k: inputs[k] for k in ["mean", "std", "norm"]})

        if self.training and (vis_training or self.vis_period > 0):
            storage = get_event_storage()
            if vis_training or storage.iter % self.vis_period == 0:
                self.visualize_training(kspace, zf_image, target, image)

        output_dict["zf_image"] = zf_image

        return output_dict

    @classmethod
    def from_config(cls, cfg):
        """
        Note:
            Currently, only resblocks can be constructed from the config.
            Step sizes are currently fixed at initialization to -2.0.
        """
        # Extract network parameters
        num_grad_steps = cfg.MODEL.UNROLLED.NUM_UNROLLED_STEPS
        share_weights = cfg.MODEL.UNROLLED.SHARE_WEIGHTS

        # Data dimensions
        num_emaps = cfg.MODEL.UNROLLED.NUM_EMAPS

        # Determine block to use for each unrolled step.
        if cfg.MODEL.UNROLLED.BLOCK_ARCHITECTURE == "ResNet":
            builder = lambda: _build_resblock(cfg)  # noqa: E731
        else:
            # TODO: Fix any inconsistencies between config's IN_CHANNELS
            # and the number of channels that the unrolled net expects.
            mcfg = cfg.clone().defrost()
            mcfg.MODEL.META_ARCHITECTURE = cfg.MODEL.UNROLLED.BLOCK_ARCHITECTURE
            mcfg = mcfg.freeze()
            builder = lambda: build_model(mcfg)  # noqa: E731

        # Declare ResNets and RNNs for each unrolled iteration
        if share_weights:
            blocks = nn.ModuleList([builder()] * num_grad_steps)
        else:
            blocks = nn.ModuleList([builder() for _ in range(num_grad_steps)])

        return {
            "blocks": blocks,
            "fix_step_size": cfg.MODEL.UNROLLED.FIX_STEP_SIZE,
            "num_emaps": num_emaps,
            "vis_period": cfg.VIS_PERIOD,
        }


def _build_resblock(cfg):
    """Build the resblock for unrolled network.

    Args:
        cfg (CfgNode): The network configuration.

    Note:
        This is a temporary method used as a base case for building
        unrolled networks with the default resblocks. In the future,
        this will be handled by :func:`meddlr.modeling.meta_arch.build_model`.
    """
    # Data dimensions
    num_emaps = cfg.MODEL.UNROLLED.NUM_EMAPS

    # ResNet parameters
    kernel_size = cfg.MODEL.UNROLLED.KERNEL_SIZE
    if len(kernel_size) == 1:
        kernel_size = kernel_size[0]
    resnet_params = dict(
        num_resblocks=cfg.MODEL.UNROLLED.NUM_RESBLOCKS,
        in_chans=2 * num_emaps,  # complex -> real/imag
        chans=cfg.MODEL.UNROLLED.NUM_FEATURES,
        kernel_size=kernel_size,
        drop_prob=cfg.MODEL.UNROLLED.DROPOUT,
        circular_pad=cfg.MODEL.UNROLLED.PADDING == "circular",
        act_type=cfg.MODEL.UNROLLED.CONV_BLOCK.ACTIVATION,
        norm_type=cfg.MODEL.UNROLLED.CONV_BLOCK.NORM,
        norm_affine=cfg.MODEL.UNROLLED.CONV_BLOCK.NORM_AFFINE,
        order=cfg.MODEL.UNROLLED.CONV_BLOCK.ORDER,
    )

    return ResNet(**resnet_params)



import logging

import torch
import torchvision.utils as tv_utils
from torch import nn

from meddlr.config.config import configurable
from meddlr.modeling.meta_arch.build import META_ARCH_REGISTRY, build_model
from meddlr.ops import complex as cplx
from meddlr.transforms.builtin.mri import MRIReconAugmentor
from meddlr.utils.events import get_event_storage
from meddlr.utils.general import move_to_device


@META_ARCH_REGISTRY.register()
class VortexModel(nn.Module):
    """VORTEX model.

    This is the generalized model implementation for augmentation-based consistency.
    It differs from :class:`N2RModel` and :class:`M2RModel` in some ways:

        1. **Generalizable augmentor**: :class:`MRIReconAugmentor` is used to
           perform augmentations.
        2. **Faster augmentations:** Augmentations are performed on the
           operating device (e.g. GPU) with large, but reproducible seeds.
        3. **Spatial augmentations**: Consistency with spatial augmentations
           are also supported. These augmentation are also used to transform
           the target image.
    """

    _version = 1
    _aliases = ["A2RModel"]

    @configurable
    def __init__(
        self,
        model: nn.Module,
        augmentor: MRIReconAugmentor,
        use_supervised_consistency: bool = False,
        vis_period: int = -1,
    ):
        """
        Args:
            model (nn.Module): The base model.
            augmentor (MRIReconAugmentor): The augmentation module.
            use_supervised_consistency (bool, optional): If ``True``, use consistency
                with supervised examples too.
            vis_period (int, optional): The period over which to visualize images.
                If ``<=0``, it is ignored. Note if the ``model`` has a ``vis_period``
                attribute, it will be overridden so that this class handles visualization.
        """
        super().__init__()

        self.model = model
        self.augmentor = augmentor
        self.use_base_grad = False  # Keep gradient for base images in transform.
        self.use_supervised_consistency = use_supervised_consistency

        # Visualization done by this model
        if hasattr(model, "vis_period") and vis_period > 0:
            self.model.vis_period = -1
        self.vis_period = vis_period

    def augment(self, inputs, pred_base):
        inputs = move_to_device(inputs, device="cuda")
        pred_base = move_to_device(pred_base, device="cuda")
        kspace, maps = inputs["kspace"].clone(), inputs["maps"].clone()

        out, _, _ = self.augmentor(kspace, maps, pred_base, mask=True)

        inputs = {
            k: v.clone() if isinstance(v, torch.Tensor) else v
            for k, v in inputs.items()
            if k not in ("kspace", "maps")
        }
        inputs["kspace"] = out["kspace"]
        inputs["maps"] = out["maps"]
        aug_pred_base = out["target"]
        return inputs, aug_pred_base

    def visualize_aug_training(self, kspace, kspace_aug, preds, preds_base, target=None):
        """Visualize training of augmented data.

        Args:
            kspace: The base kspace.
            kspace_aug: The augmented kspace.
            preds: Reconstruction of augmented kspace. Shape: NxHxWx2.
            preds_base: Reconstruction of base kspace. Shape: NxHxWx2.
        """
        storage = get_event_storage()

        with torch.no_grad():
            # calc mask for first coil only
            if cplx.is_complex(kspace):
                kspace = torch.view_as_real(kspace)
            kspace = kspace.cpu()[0, ..., 0, :].unsqueeze(0)
            if cplx.is_complex(kspace_aug):
                kspace_aug = torch.view_as_real(kspace_aug)
            kspace_aug = kspace_aug.cpu()[0, ..., 0, :].unsqueeze(0)
            preds = preds.cpu()[0, ...].unsqueeze(0)
            preds_base = preds_base.cpu()[0, ...].unsqueeze(0)

            all_images = [preds, preds_base]
            errors = [cplx.abs(preds_base - preds)]
            if target is not None:
                target = target.cpu()[0, ...].unsqueeze(0)
                all_images.append(target)
                errors.append(cplx.abs(target - preds))

            all_images = torch.cat(all_images, dim=2)
            all_kspace = torch.cat([kspace, kspace_aug], dim=2)
            errors = torch.cat(errors, dim=2)

            imgs_to_write = {
                "phases": cplx.angle(all_images),
                "images": cplx.abs(all_images),
                "errors": errors,
                "masks": cplx.get_mask(kspace),
                "kspace": cplx.abs(all_kspace),
            }

            for name, data in imgs_to_write.items():
                data = data.squeeze(-1).unsqueeze(1)
                data = tv_utils.make_grid(data, nrow=1, padding=1, normalize=True, scale_each=True)
                storage.put_image("train_aug/{}".format(name), data.numpy(), data_format="CHW")

    def forward(self, inputs):
        if not self.training:
            assert (
                "unsupervised" not in inputs
            ), "unsupervised inputs should not be provided in eval mode"
            inputs = inputs.get("supervised", inputs)
            return self.model(inputs)

        vis_training = False
        if self.training and self.vis_period > 0:
            storage = get_event_storage()
            if storage.iter % self.vis_period == 0:
                vis_training = True

        inputs_supervised = inputs.get("supervised", None)
        inputs_unsupervised = inputs.get("unsupervised", None)
        if inputs_supervised is None and inputs_unsupervised is None:
            raise ValueError("Examples not formatted in the proper way")
        output_dict = {}

        # Recon
        if inputs_supervised is not None:
            output_dict["recon"] = self.model(
                inputs_supervised, return_pp=True, vis_training=vis_training
            )

        # Consistency.
        # kspace_aug = kspace + U \sigma \mathcal{N}
        # Loss = L(f(Ti(Te(kspace)), \theta), Te(f(kspace, \theta)))
        inputs_consistency = []
        if inputs_unsupervised is not None:
            inputs_consistency.append(inputs_unsupervised)
        if self.use_supervised_consistency and inputs_supervised is not None:
            inputs_consistency.append({k: v for k, v in inputs_supervised.items() if k != "target"})

        if len(inputs_consistency) > 0:
            if len(inputs_consistency) > 1:
                inputs_consistency = {
                    k: torch.cat([x[k] for x in inputs_consistency], dim=0)
                    for k in inputs_consistency[0].keys()
                }
            else:
                inputs_consistency = inputs_consistency[0]
            with torch.no_grad():
                pred_base = self.model(inputs_consistency)
                # Target only used for visualization purposes not for loss.
                target = inputs_unsupervised.get("target", None)
                pred_base = pred_base["pred"]
            inputs_consistency_aug, pred_base = self.augment(inputs_consistency, pred_base)
            pred_aug = self.model(inputs_consistency_aug, return_pp=True)
            if "target" in pred_aug:
                del pred_aug["target"]
            pred_aug["target"] = pred_base.detach()
            output_dict["consistency"] = pred_aug
            if vis_training:
                self.visualize_aug_training(
                    inputs_consistency["kspace"],
                    inputs_consistency_aug["kspace"],
                    pred_aug["pred"],
                    pred_base,
                    target=target,
                )

        return output_dict

    @classmethod
    def from_config(cls, cfg):
        _logger = logging.getLogger(__name__)

        model_cfg = cfg.clone()
        model_cfg.defrost()
        model_cfg.MODEL.META_ARCHITECTURE = cfg.MODEL.A2R.META_ARCHITECTURE
        model_cfg.freeze()
        model = build_model(model_cfg)

        augmentor = MRIReconAugmentor.from_cfg(
            cfg, aug_kind="consistency", device=cfg.MODEL.DEVICE, seed=cfg.SEED
        )
        _logger.info("Built augmentor:\n{}".format(str(augmentor.tfms_or_gens)))

        return {
            "model": model,
            "augmentor": augmentor,
            "use_supervised_consistency": cfg.MODEL.A2R.USE_SUPERVISED_CONSISTENCY,
            "vis_period": cfg.VIS_PERIOD,
        }



import logging
from collections import defaultdict
from typing import Any, Dict, List, Optional

import numpy as np
from tqdm import tqdm

from meddlr.data.data_utils import structure_patches
from meddlr.evaluation.evaluator import DatasetEvaluator

_logger = logging.getLogger(__name__)


class ScanEvaluator(DatasetEvaluator):
    _remaining_state: Optional[Dict[str, Any]]
    _predictions: List[Dict[str, Any]]
    _is_flushing: bool
    _logger: logging.Logger

    def enter_prediction_scope(self, skip_last_scan: bool = True):
        if self._remaining_state is not None:
            raise ValueError(
                "You must exit the prediction scope by calling `exit_prediction_scope`."
            )

        remaining_preds = []
        if skip_last_scan:
            try:
                scan_ids = np.asarray([p["metadata"]["scan_id"] for p in self._predictions])
            except KeyError:
                raise ValueError(
                    "Cannot skip last scan. metadata does not contain 'scan_id' keyword."
                )

            change_idxs = np.where(scan_ids[1:] != scan_ids[:-1])[0]
            if len(change_idxs) == 0:
                self._logger.warning(
                    "Flushing skipped. All current predictions are from the same scan. "
                    "To force flush, set `skip_last_scan=False`."
                )
                return False

            last_idx = int(change_idxs[-1] + 1)
            remaining_preds = self._predictions[last_idx:]
            self._predictions = self._predictions[:last_idx]

        self._remaining_state = {"_predictions": remaining_preds}

    def exit_prediction_scope(self):
        if self._remaining_state is None:
            raise ValueError(
                "You must enter prediction scope first by calling `enter_prediction_scope`"
            )

        remaining_state = {k: v for k, v in self._remaining_state.items() if hasattr(self, k)}
        for k, v in remaining_state.items():
            setattr(self, k, v)
        self._remaining_state = None

    def flush(self, enter_prediction_scope: bool = True, skip_last_scan: bool = True):
        """Clear predictions and computing running metrics.

        Results are added to ``self._running_results``.

        Args:
            enter_prediction_scope (bool, optional): If ``True``, enter the
                prediction scope.
            skip_last_scan (bool, optional): If ``True``, does not flush
                most recent scan. This avoids prematurely computing metrics
                before all slices of the scan are available.
        """
        if enter_prediction_scope:
            success = self.enter_prediction_scope(skip_last_scan=skip_last_scan)
            # TODO: this means we haven't properly set the state, which is likely because
            # we dont want to execute this flush. Make the API clearer.
            if success is False:
                return

        self._logger.info("Flushing results...")

        self._is_flushing = True
        self.evaluate()
        self._is_flushing = False

        if enter_prediction_scope:
            self.exit_prediction_scope()

    def structure_scans(self, verbose=True):
        """Structure scans into volumes to be used to evaluation."""
        return structure_scans(self._predictions, verbose=verbose)


def structure_scans(
    predictions,
    to_struct=("pred", "target"),
    dims: Dict[int, str] = None,
    metadata=("voxel_spacing", "affine"),
    contiguous: bool = False,
    verbose: bool = True,
) -> Dict[str, Dict[str, Any]]:
    """Structure patches into single objects.

    Note:
        This may not currently be a very fast implementation as stacking occurs multiple times
        instead of single tensor allocation.

    TODO: Benchmark speeds for future releases.
    """
    if not dims:
        dims = {0: "slice_id"}

    if not bool(predictions[0]["metadata"]):
        raise ValueError("Cannot aggregate scans without metadata")
    if len(set(to_struct) & set(metadata)) > 0:
        raise ValueError(
            "Cannot overlap between keys to structure (`to_struct`) "
            "and keys for metadata (`metadata`)."
        )

    if verbose:
        _logger.info("Structuring slices into volumes...")

    scan_map = defaultdict(dict)
    for pred in predictions:
        values = tuple(pred["metadata"][k] for k in dims.values())
        scan_map[pred["metadata"]["scan_id"]][values] = pred

    if verbose:
        _logger.info("Structuring slices into volumes...")

    scans = {}
    for scan_id, pred_dict in tqdm(scan_map.items()):
        coords = list(pred_dict.keys())
        pred = {
            k: structure_patches([pred_dict[c][k] for c in coords], coords=coords, dims=dims)
            for k in to_struct
        }
        if contiguous:
            pred = {k: v.contiguous() for k, v in pred.items()}
        sample_dict = pred_dict[coords[0]]
        pred.update({k: sample_dict["metadata"].get(k, None) for k in metadata})
        scans[scan_id] = pred
    return scans



import copy
import itertools
import logging
import os
import time
from collections import defaultdict
from typing import List, Sequence, Union

import numpy as np
import pandas as pd
import silx.io.dictdump as sio
import torch
from tqdm import tqdm

import meddlr.utils.comm as comm
from meddlr.data.transforms.transform import build_normalizer
from meddlr.evaluation.scan_evaluator import ScanEvaluator, structure_scans
from meddlr.metrics.build import build_metrics
from meddlr.metrics.collection import MetricCollection
from meddlr.ops import complex as cplx


class ReconEvaluator(ScanEvaluator):
    """
    Evaluate reconstruction quality using the metrics listed below:

    - reconstruction loss (as specified by `loss_computer`)
    - L1, L2
    - Magnitude PSNR
    - Complex PSNR
    - SSIM (to be implemented)
    """

    def __init__(
        self,
        dataset_name,
        cfg,
        distributed=False,
        sync_outputs=False,
        aggregate_scans=True,
        group_by_scan=False,
        output_dir=None,
        skip_rescale=False,
        save_scans=False,
        metrics=None,
        flush_period: int = None,
        to_cpu=False,
        channel_names=None,
        eval_in_process=False,
        structure_channel_by=None,
        prefix="val",
    ):
        """
        Args:
            dataset_name (str): name of the dataset to be evaluated.
            cfg (CfgNode): config instance
            output_dir (str): optional, an output directory to dump all
                results predicted on the dataset.
            distributed (bool, optional): If ``True``, collect results from all
                ranks for evaluation. Otherwise, will evaluate the results in the
                current process. ∂If using ``DistributedDataParallel``, this should likely
                be ``True``.
            sync_outputs (bool, optional): If ``True``, synchronizes all predictions
                before evaluation. If ``False``, synchronizes metrics before reduction.
                Ignored if ``distributed=False``.
            aggregate_scans (bool, optional): If ``True``, also computes metrics per
                scan under the label `'scan_{metric}'` (e.g. scan_l1).
            group_by_scan (bool, optional): If `True`, groups metrics by scan.
                `self.evaluate()` will return a dict of scan_id -> dict[metric name, metric value]
            skip_rescale (bool, optional): If `True`, skips rescaling the output and target
                by the mean/std.
            save_scans (bool, optional): If `True`, saves predictions to .npy file.
            metrics (Sequence[str], optional): To avoid computing metrics, set to ``False``.
                Defaults to all supported recon metrics. To process metrics on the full scan,
                append ``'_scan'`` to the metric name (e.g. `'psnr_scan'`).
            flush_period (int, optional): The approximate period over which predictions
                are cleared and running results are computed. This parameter helps
                mitigate OOM errors. The period is equivalent to number of examples
                (not batches).
            to_cpu (bool, optional): If ``True``, move all data to the cpu to do computation.
            eval_in_process (bool, optional): If ``True``, run slice/patch evaluation
                while processing. This may increase overall throughput.
        """
        self._dataset_name = dataset_name
        self._output_dir = output_dir
        if self._output_dir:
            os.makedirs(self._output_dir, exist_ok=True)

        self._cpu_device = torch.device("cpu")
        self._logger = logging.getLogger(__name__)
        self._normalizer = build_normalizer(cfg)
        self._group_by_scan = group_by_scan
        self._distributed = distributed
        self._sync_outputs = sync_outputs
        self._aggregate_scans = aggregate_scans
        self._skip_rescale = skip_rescale
        self._channel_names = channel_names
        self._structure_channel_by = structure_channel_by
        self._prefix = prefix

        if save_scans and (not output_dir or not aggregate_scans):
            raise ValueError("`output_dir` and `aggregate_scans` must be specified to save scans.")
        self._save_scans = save_scans
        self._save_scan_dir = os.path.join(self._output_dir, "pred") if self._output_dir else None
        if self._save_scan_dir:
            os.makedirs(self._save_scan_dir, exist_ok=True)

        if metrics is False:
            metrics = []
        elif metrics in (None, True):
            metrics = self.default_metrics()
        self._metric_names = metrics

        self._results = None

        if flush_period is None:
            flush_period = cfg.TEST.FLUSH_PERIOD
        if distributed and flush_period != 0:
            raise ValueError("Result flushing is not enabled in distributed mode.")
        self.flush_period = flush_period
        self.to_cpu = to_cpu
        self.eval_in_process = eval_in_process

        self._remaining_state = None
        self._predictions = []
        self._is_flushing = False

        # Memory
        self._memory = defaultdict(list)

    @classmethod
    def default_metrics(cls) -> List[str]:
        """The default metrics processed by this class."""
        metrics = ["psnr", "psnr_mag", "ssim (Wang)", "nrmse", "nrmse_mag"]
        metrics.extend([f"{x}_scan" for x in metrics])
        return metrics

    def reset(self):
        self._remaining_state = None
        self._predictions = []
        self._is_flushing = False
        self._memory = defaultdict(list)

        metrics = self._metric_names
        prefix = self._prefix + "_" if self._prefix else ""
        slice_metrics = [m for m in metrics if not m.endswith("_scan")]
        scan_metrics = [m[:-5] for m in metrics if m.endswith("_scan")]
        self.slice_metrics = build_metrics(
            slice_metrics,
            fmt=prefix + "{}",
            channel_names=self._channel_names,
        )
        self.scan_metrics = build_metrics(
            scan_metrics,
            fmt=prefix + "{}_scan",
            channel_names=self._channel_names,
        )
        self.slice_metrics.eval()
        self.scan_metrics.eval()

    def exit_prediction_scope(self):
        ret_val = super().exit_prediction_scope()
        if torch.cuda.is_available() and torch.cuda.is_initialized():
            torch.cuda.empty_cache()
        return ret_val

    def process(self, inputs, outputs):
        """
        Args:
            inputs: the inputs to a recon model (e.g., GeneralizedRCNN).
                Currently this should be an empty dictionary.
            outputs: the outputs of a COCO model. It is a list of dicts with key
                "instances" that contains :class:`Instances`.

        Note:
            All elements in ``inputs`` and ``outputs`` should already
            be detached from the computational graph.
        """
        N = outputs["pred"].shape[0]

        preds: torch.Tensor
        targets: torch.Tensor

        if self._skip_rescale:
            # Do not rescale the outputs
            preds = outputs["pred"]
            targets = outputs["target"]
        else:
            normalized = self._normalizer.undo(
                image=outputs["pred"],
                target=outputs["target"],
                mean=inputs["mean"],
                std=inputs["std"],
                channels_last=True,
            )
            preds = normalized["image"]
            targets = normalized["target"]

        if self.to_cpu:
            preds = preds.to(self._cpu_device, non_blocking=True)
            targets = targets.to(self._cpu_device, non_blocking=True)

        if self.eval_in_process:
            self.evaluate_prediction(
                {"pred": preds, "target": targets},
                self.slice_metrics,
                [
                    "-".join([str(md[field]) for field in ("scan_id", "slice_id")])
                    for md in inputs["metadata"]
                ],
                is_batch=True,
            )

        self._predictions.extend(
            [
                {
                    "pred": preds[i],
                    "target": targets[i],
                    "metadata": inputs["metadata"][i] if "metadata" in inputs else {},
                }
                for i in range(N)
            ]
        )
        self._append_memory("after_predictions")

        has_num_examples = self.flush_period > 0 and len(self._predictions) >= self.flush_period
        has_num_scans = self.flush_period < 0 and len(
            {x["metadata"]["scan_id"] for x in self._predictions}
        ) > abs(self.flush_period)
        if has_num_examples or has_num_scans:
            self.flush(enter_prediction_scope=True, skip_last_scan=True)
            self._append_memory("after_flush")

    def structure_scans(self, verbose=True):
        """Structure scans into volumes to be used to evaluation."""
        structure_channel_by = self._structure_channel_by
        structure_by = {0: "slice_id"}
        if structure_channel_by is not None:
            # This does not work when predictions are real/imaginary are separate channels
            # TODO: Fix this.
            structure_by[-1] = structure_channel_by
        to_struct = ("pred", "target")

        # Making a tensor contiguous can be an expensive operation.
        # We want to do it as few times as possible. Because we have to
        # do it anyway when we squeeze the tensor when structuring by channel,
        # we opt not to do when first structuring the scans.
        contiguous = structure_channel_by is None

        out = structure_scans(
            self._predictions,
            to_struct=to_struct,
            dims=structure_by,
            contiguous=contiguous,
            verbose=verbose,
        )

        if structure_channel_by is not None:
            for scan_id in out:
                out[scan_id].update(
                    {k: out[scan_id][k].squeeze(-2).contiguous() for k in to_struct}
                )

        return out

    def synchronize_predictions(self):
        comm.synchronize()
        self._predictions = comm.gather(self._predictions, dst=0)
        self._predictions = list(itertools.chain(*self._predictions))
        if not comm.is_main_process():
            self._predictions = []

    def evaluate(self):
        # Sync predictions (if applicable)
        if self._distributed and self._sync_outputs:
            self.synchronize_predictions()
            if not self._predictions:
                return {}

        if len(self._predictions) == 0:
            self._logger.warning("[ReconEvaluator] Did not receive valid predictions.")
            return {}

        # Compute metrics per slice (if not already done during process step).
        if not self.eval_in_process:
            for pred in tqdm(
                self._predictions, desc="Slice metric", disable=not comm.is_main_process()
            ):
                self.evaluate_prediction(
                    pred,
                    self.slice_metrics,
                    "-".join([str(pred["metadata"][x]) for x in ("scan_id", "slice_id")]),
                )

        # Compute metrics per scan.
        has_metadata = bool(self._predictions[0]["metadata"])
        if self._aggregate_scans and has_metadata:
            scans = self.structure_scans()
            for scan_id, pred in tqdm(
                scans.items(), desc="Scan Metrics", disable=not comm.is_main_process()
            ):
                self.evaluate_prediction(pred, self.scan_metrics, scan_id)
                if self._save_scans:
                    sio.dicttoh5(
                        {"pred": pred["pred"].cpu()},
                        os.path.join(self._save_scan_dir, f"{scan_id}.h5"),
                    )

        if self._group_by_scan:
            pred_vals = self._group_results_by_scan()
        else:
            pred_vals = self.slice_metrics.to_dict()
            pred_vals.update(self.scan_metrics.to_dict())

        self._results = pred_vals

        if not self._is_flushing:
            self.log_summary()

        # Copy so the caller can do whatever with results
        return copy.deepcopy(self._results)

    def _group_results_by_scan(self):
        """Get results grouping by scan."""
        pred_vals = defaultdict(dict)

        slice_metrics = self.slice_metrics.to_dict(group_by=["id", "Metric"])
        agg_slice_metrics = defaultdict(lambda: defaultdict(list))
        for (slice_id, metric_name), value in slice_metrics.items():
            agg_slice_metrics[slice_id.rsplit("-", 1)[0]][metric_name].append(value)
        for scan_id, metrics_dict in agg_slice_metrics.items():
            for metrics_name, values in metrics_dict.items():
                pred_vals[scan_id][metrics_name] = np.mean(values)

        scan_metrics = self.scan_metrics.to_dict(group_by=["id", "Metric"])
        for (scan_id, metric_name), value in scan_metrics.items():
            pred_vals[scan_id][metric_name] = value

        return pred_vals

    def log_summary(self):
        if not comm.is_main_process():
            return

        output_dir = self._output_dir
        self._logger.info(
            "[{}] Slice metrics summary:\n{}".format(
                type(self).__name__, self.slice_metrics.summary()
            )
        )
        # TODO: Make this based off if metrics has scans
        if self._aggregate_scans:
            self._logger.info(
                "[{}] Scan metrics summary:\n{}".format(
                    type(self).__name__, self.scan_metrics.summary()
                )
            )

        if not output_dir:
            return

        dirpath = output_dir
        os.makedirs(dirpath, exist_ok=True)
        test_results_summary_path = os.path.join(dirpath, "results.txt")
        slice_metrics_path = os.path.join(dirpath, "slice_metrics.csv")
        scan_metrics_path = os.path.join(dirpath, "scan_metrics.csv")

        # Write details to test file
        with open(test_results_summary_path, "w+") as f:
            f.write("Results generated on %s\n" % time.strftime("%X %x %Z"))
            # f.write("Weights Loaded: %s\n" % os.path.basename(self._config.TEST_WEIGHT_PATH))

            f.write("--" * 40)
            f.write("\n")
            f.write("Slice Metrics:\n")
            f.write(self.slice_metrics.summary())
            f.write("--" * 40)
            f.write("\n")
            if self._aggregate_scans:
                f.write("Scan Metrics:\n")
                f.write(self.scan_metrics.summary())
            f.write("--" * 40)
            f.write("\n")

        df: pd.DataFrame = self.slice_metrics.to_pandas()
        df.to_csv(slice_metrics_path, header=True, index=True)

        df: pd.DataFrame = self.scan_metrics.to_pandas()
        df.to_csv(scan_metrics_path, header=True, index=True)

    def evaluate_prediction(
        self,
        prediction,
        metrics: MetricCollection,
        ex_id: Union[str, Sequence[str]],
        is_batch=False,
    ):
        output, target = prediction["pred"], prediction["target"]
        if not is_batch:
            output, target = output.unsqueeze(0), target.unsqueeze(0)
            ex_id = [ex_id]
        output, target = cplx.channel_first(output), cplx.channel_first(target)
        metrics(preds=output, targets=target, ids=ex_id)
        return metrics.to_dict()

    def _append_memory(self, key):
        if not torch.cuda.is_available():
            return
        self._memory[key].append(torch.cuda.max_memory_allocated() / 1024.0 / 1024.0)
        mem = self._memory[key]
        if len(mem) > 1 and (mem[-1] - mem[-2] > 500):
            self._logger.info(f"Memory exceeded '{key}'- previous 5 logs: {mem[-5:]}")
            # self._logger.info(torch.cuda.memory_stats())



import os
from typing import List, Mapping

import torch

from meddlr.utils import env

_PATH_MANAGER = env.get_path_manager()


def move_to_device(obj, device, non_blocking=False, base_types=None):
    """Given a structure (possibly) containing Tensors on the CPU, move all the Tensors
      to the specified GPU (or do nothing, if they should be on the CPU).
        device = -1 -> "cpu"
        device =  0 -> "cuda:0"

    Args:
      obj(Any): The object to convert.
      device(int): The device id, defaults to -1.

    Returns:
      Any: The converted object.
    """
    if not torch.cuda.is_available() or (isinstance(device, int) and device < 0):
        device = "cpu"
    elif isinstance(device, int):
        device = f"cuda:{device}"

    if isinstance(obj, torch.Tensor):
        return obj.to(device, non_blocking=non_blocking)  # type: ignore
    elif base_types is not None and isinstance(obj, base_types):
        return obj.to(device)
    elif isinstance(obj, dict):
        return {
            key: move_to_device(value, device, non_blocking=non_blocking, base_types=base_types)
            for key, value in obj.items()
        }
    elif isinstance(obj, list):
        return [
            move_to_device(item, device, non_blocking=non_blocking, base_types=base_types)
            for item in obj
        ]
    elif isinstance(obj, tuple):
        return tuple(
            [
                move_to_device(item, device, non_blocking=non_blocking, base_types=base_types)
                for item in obj
            ]
        )
    else:
        return obj


def find_experiment_dirs(dirpath, completed=True) -> List[str]:
    """Recursively search for experiment directories under the ``dirpath``.

    Args:
        dirpath (str): The base directory under which to search.
        completed (bool, optional): If `True`, filter directories where runs
            are completed.

    Returns:
        exp_dirs (List[str]): A list of experiment directories.
    """

    def _find_exp_dirs(_dirpath):
        # Directories with "config.yaml" are considered experiment directories.
        if os.path.isfile(os.path.join(_dirpath, "config.yaml")):
            return [_dirpath]
        # Directories with no more subdirectories do not have a path.
        subfiles = [os.path.join(_dirpath, x) for x in os.listdir(_dirpath)]
        subdirs = [x for x in subfiles if os.path.isdir(x)]
        if len(subdirs) == 0:
            return []
        exp_dirs = []
        for dp in subdirs:
            exp_dirs.extend(_find_exp_dirs(dp))
        return exp_dirs

    dirpath = _PATH_MANAGER.get_local_path(dirpath)
    exp_dirs = _find_exp_dirs(dirpath)
    if completed:
        exp_dirs = [x for x in exp_dirs if os.path.isfile(os.path.join(x, "model_final.pth"))]
    return exp_dirs


def flatten_dict(results, delimiter="/"):
    """
    Expand a hierarchical dict of scalars into a flat dict of scalars.
    If results[k1][k2][k3] = v, the returned dict will have the entry
    {"k1/k2/k3": v}.

    Args:
        results (dict):
    """
    r = {}
    for k, v in results.items():
        if isinstance(v, Mapping):
            v = flatten_dict(v)
            for kk, vv in v.items():
                r[k + delimiter + kk] = vv
        else:
            r[k] = v
    return r



from typing import Dict, List

from torch.utils.data.dataloader import default_collate as _default_collate

__all__ = ["default_collate", "collate_by_supervision"]


def default_collate(batch: List[Dict]):
    """Default collate function for ss_recon.

    This collate function handles metadata appropriately by returning
    metadata as a list of dictionaries instead of a dictionary of tensors.
    This is done because not all metadata (e.g. string values) can be
    tensorized.

    Metadata is only handled if at least one example in the batch has
    a ``'metadata'`` key.

    Args:
        batch (list): The list of dictionaries.

    Returns:
        Dict
    """
    metadata = None
    if any("metadata" in b for b in batch):
        metadata = [b.pop("metadata", None) for b in batch]
    out_dict = _default_collate(batch)
    if metadata is not None:
        out_dict["metadata"] = metadata
    return out_dict


def collate_by_supervision(batch: list):
    """Collate supervised/unsupervised batch examples.

    This collate function is required when training with semi-supervised
    models, such as :cls:`N2RModel` and :cls:`VortexModel`.

    Args:
        batch (list): The list of dictionaries.

    Returns:
        Dict[str, Dict]: A dictionary with 2 keys, ``'supervised'`` and ``'unsupervised'``.
    """
    supervised = [x for x in batch if not x.get("is_unsupervised", False)]
    unsupervised = [x for x in batch if x.get("is_unsupervised", False)]

    out_dict = {}
    if len(supervised) > 0:
        supervised = default_collate(supervised)
        out_dict["supervised"] = supervised
    if len(unsupervised) > 0:
        unsupervised = default_collate(unsupervised)
        out_dict["unsupervised"] = unsupervised
    assert len(out_dict) > 0
    return out_dict



import functools
from typing import Optional, Sequence

from fvcore.common.registry import Registry as _Registry
from tabulate import tabulate


class Registry(_Registry):
    """Extension of fvcore's registry that supports aliases."""

    _ALIAS_KEYWORDS = ("_aliases", "_ALIASES")

    def __init__(self, name: str):
        super().__init__(name=name)
        self._metadata_map = {}

    def _get_aliases(self, obj_func_or_class):
        for kw in self._ALIAS_KEYWORDS:
            if hasattr(obj_func_or_class, kw):
                return getattr(obj_func_or_class, kw)
        return []

    def register(self, obj: object = None, aliases: Sequence[str] = None) -> Optional[object]:
        if obj is None:
            # used as a decorator
            def deco(func_or_class: object, aliases=None) -> object:
                name = func_or_class.__name__
                self._do_register(name, func_or_class)
                if aliases is None:
                    aliases = self._get_aliases(func_or_class)
                if not isinstance(aliases, (list, tuple, set)):
                    aliases = [aliases]
                for alias in aliases:
                    self._do_register(alias, func_or_class, is_alias=True)
                return func_or_class

            kwargs = {"aliases": aliases}
            if any(v is not None for v in kwargs.values()):
                return functools.partial(deco, **kwargs)
            else:
                return deco

        name = obj.__name__
        self._do_register(name, obj)
        if aliases is None:
            aliases = self._get_aliases(obj) if isinstance(obj, type) else []
        for alias in aliases:
            self._do_register(alias, obj, is_alias=True)

    def _do_register(self, name: str, obj: object, **kwargs) -> None:
        docstring = obj.__doc__
        if docstring is None:
            docstring = ""

        aliases = self._get_aliases(obj) if isinstance(obj, type) else None
        if not aliases:
            aliases = None

        self._metadata_map[name] = {
            "name": name,
            "description": kwargs.pop("description", docstring.split("\n")[0]),
            "aliases": aliases,
            **kwargs,
        }
        return super()._do_register(name, obj)

    def clear(self):
        self._obj_map = {}
        self._metadata_map = {}

    def __repr__(self) -> str:
        metadata = [v for v in self._metadata_map.values() if not v.get("is_alias", False)]
        table = tabulate(metadata, headers="keys", tablefmt="fancy_grid")
        return "Registry of {}:\n{}".format(self._name, table)



from abc import ABC, abstractmethod

import torch
from fvcore.common.registry import Registry

from meddlr.data.transforms.transform import build_normalizer
from meddlr.ops import complex as cplx
from meddlr.utils import transforms as T

LOSS_COMPUTER_REGISTRY = Registry("LOSS_COMPUTER")  # noqa F401 isort:skip
LOSS_COMPUTER_REGISTRY.__doc__ = """
Registry for loss computers.

The registered object will be called with `obj(cfg)`
and expected to return a LossComputer object.
"""

EPS = 1e-11
IMAGE_LOSSES = ["l1", "l2", "psnr", "nrmse", "mag_l1", "perp_loss"]
KSPACE_LOSSES = ["k_l1", "k_l1_normalized", "k_l1_l2_sum_normalized"]


def build_loss_computer(cfg, name, **kwargs):
    return LOSS_COMPUTER_REGISTRY.get(name)(cfg, **kwargs)


class LossComputer(ABC):
    def __init__(self, cfg):
        self._normalizer = build_normalizer(cfg)

    @abstractmethod
    def __call__(self, input, output):
        pass

    def _get_metrics(self, target: torch.Tensor, output: torch.Tensor, loss_name):
        # Compute metrics
        if loss_name == "mag_l1":
            abs_error = torch.abs(output - target)
            abs_mag_error = abs_error
        else:
            abs_error = cplx.abs(output - target)
            abs_mag_error = torch.abs(cplx.abs(output) - cplx.abs(target))
        l1 = torch.mean(abs_error)
        mag_l1 = torch.mean(abs_mag_error)
        N = target.shape[0]

        abs_error = abs_error.view(N, -1)
        if loss_name == "mag_l1":
            tgt_mag = torch.abs(target).view(N, -1)
        else:
            tgt_mag = cplx.abs(target).view(N, -1)
        l2 = torch.sqrt(torch.mean(abs_error ** 2, dim=1))
        psnr = 20 * torch.log10(tgt_mag.max(dim=1)[0] / (l2 + EPS))
        nrmse = l2 / torch.sqrt(torch.mean(tgt_mag ** 2, dim=1))

        metrics_dict = {
            "l1": l1,
            "l2": l2.mean(),
            "psnr": psnr.mean(),
            "nrmse": nrmse.mean(),
            "mag_l1": mag_l1,
        }

        if loss_name == "perp_loss":
            metrics_dict.update(perp_loss(output, target))

        if loss_name in KSPACE_LOSSES:
            target, output = T.fft2(target), T.fft2(output)
            abs_error = cplx.abs(target - output)
            if loss_name == "k_l1":
                metrics_dict["loss"] = torch.mean(abs_error)
            elif loss_name == "k_l1_normalized":
                metrics_dict["loss"] = torch.mean(abs_error / (cplx.abs(target) + EPS))
            elif loss_name == "k_l1_l2_sum_normalized":
                kl1_norm = torch.sum(abs_error) / torch.sum(cplx.abs(target))
                kl2_norm = torch.sqrt(torch.sum(abs_error ** 2)) / torch.sqrt(
                    torch.sum(cplx.abs(target) ** 2)
                )  # noqa: E501
                metrics_dict["loss"] = 0.5 * kl1_norm + 0.5 * kl2_norm
            else:
                assert False  # should not reach here
        else:
            loss = metrics_dict[loss_name]
            metrics_dict["loss"] = loss

        return metrics_dict


@LOSS_COMPUTER_REGISTRY.register()
class BasicLossComputer(LossComputer):
    def __init__(self, cfg):
        super().__init__(cfg)
        loss_name = cfg.MODEL.RECON_LOSS.NAME
        assert loss_name in IMAGE_LOSSES or loss_name in KSPACE_LOSSES
        self.loss = loss_name
        self.renormalize_data = cfg.MODEL.RECON_LOSS.RENORMALIZE_DATA

    def __call__(self, input, output):
        pred: torch.Tensor = output["pred"]
        target = output["target"].to(pred.device)

        if self.renormalize_data:
            normalization_args = {k: input.get(k, output[k]) for k in ["mean", "std"]}
            normalized = self._normalizer.undo(
                image=pred,
                target=target,
                mean=normalization_args["mean"],
                std=normalization_args["std"],
            )
            output = normalized["image"]
            target = normalized["target"]
        else:
            output = pred

        metrics_dict = self._get_metrics(target, output, self.loss)
        return metrics_dict


@LOSS_COMPUTER_REGISTRY.register()
class N2RLossComputer(LossComputer):
    def __init__(self, cfg):
        super().__init__(cfg)
        recon_loss = cfg.MODEL.RECON_LOSS.NAME
        consistency_loss = cfg.MODEL.CONSISTENCY.LOSS_NAME
        latent_loss = cfg.MODEL.CONSISTENCY.LATENT_LOSS_NAME

        assert recon_loss in IMAGE_LOSSES or recon_loss in KSPACE_LOSSES
        assert consistency_loss in IMAGE_LOSSES or consistency_loss in KSPACE_LOSSES

        self.recon_loss = recon_loss
        self.consistency_loss = consistency_loss
        self.latent_loss = latent_loss
        self.renormalize_data = cfg.MODEL.RECON_LOSS.RENORMALIZE_DATA
        self.consistency_weight = cfg.MODEL.CONSISTENCY.LOSS_WEIGHT
        self.latent_weight = cfg.MODEL.CONSISTENCY.LATENT_LOSS_WEIGHT
        self.use_latent = cfg.MODEL.CONSISTENCY.USE_LATENT
        self.use_consistency = cfg.MODEL.CONSISTENCY.USE_CONSISTENCY
        self.num_latent_layers = cfg.MODEL.CONSISTENCY.NUM_LATENT_LAYERS
        self.latent_keys = ["E4", "E3", "D3", "E2", "D2", "E1", "D1"]
        # self.use_robust = cfg.MODEL.LOSS.USE_ROBUST
        # self.beta = cfg.MODEL.LOSS.BETA
        # self.robust_step_size = cfg.MODEL.LOSS.ROBUST_STEP_SIZE

    def _compute_metrics(self, input, output, loss):
        """Computes image metrics on prediction and target data."""
        if output is None or len(output) == 0:
            return {k: torch.Tensor([0.0]).detach() for k in ["l1", "l2", "psnr", "loss"]}

        pred: torch.Tensor = output["pred"]
        target = output["target"].to(pred.device)
        if self.renormalize_data:
            normalized = self._normalizer.undo(
                image=pred, target=target, mean=input["mean"], std=input["std"]
            )
            output = normalized["image"]
            target = normalized["target"]
        else:
            output = pred

        # Compute metrics
        metrics_dict = self._get_metrics(target, output, loss)
        return metrics_dict

    # def compute_robust_loss(self, group_loss):
    #     if torch.is_grad_enabled():  # update adv_probs if in training mode
    #         adjusted_loss = group_loss
    #         if self.do_adj:
    #             adjusted_loss += self.loss_adjustment
    #         logit_step = self.robust_step_size * adjusted_loss.data
    #         if self.stable:
    #             self.adv_probs_logits = self.adv_probs_logits + logit_step
    #         else:
    #             self.adv_probs = self.adv_probs * torch.exp(logit_step)
    #             self.adv_probs = self.adv_probs / self.adv_probs.sum()
    #
    #     if self.stable:
    #         adv_probs = torch.softmax(self.adv_probs_logits, dim=-1)
    #     else:
    #         adv_probs = self.adv_probs
    #     robust_loss = group_loss @ adv_probs
    #     return robust_loss, adv_probs

    def __call__(self, input, output):
        output_recon = output.get("recon", None)
        output_consistency = output.get("consistency", None)

        loss = 0
        metrics_recon = {
            "recon_{}".format(k): v
            for k, v in self._compute_metrics(
                input.get("supervised", None), output_recon, self.recon_loss
            ).items()
        }
        if output_recon is not None:
            loss += metrics_recon["recon_loss"]

        metrics_consistency = {
            "cons_{}".format(k): v
            for k, v in self._compute_metrics(
                input.get("unsupervised", None), output_consistency, self.consistency_loss
            ).items()  # noqa
        }
        if output_consistency is not None and self.use_consistency:
            loss += self.consistency_weight * metrics_consistency["cons_loss"]

        if self.use_latent:
            num_losses = self.num_latent_layers * 2 - 1
            all_metrics_latent = []
            for i in range(num_losses):
                output_latent = {}
                output_latent["target"] = output_recon["latent"][self.latent_keys[i]]
                output_latent["pred"] = output_consistency["latent"][self.latent_keys[i]]

                metrics_latent = {
                    "latent_" + self.latent_keys[i] + "_{}".format(k): v
                    for k, v in self._compute_metrics(
                        None, output_latent, self.latent_loss
                    ).items()  # noqa
                }

                all_metrics_latent.append(metrics_latent)
                loss += (
                    self.latent_weight * metrics_latent["latent_" + self.latent_keys[i] + "_loss"]
                )

        metrics = {}
        if output_consistency is not None and self.use_consistency:
            metrics.update(metrics_consistency)
        if output_recon is not None:
            metrics.update(metrics_recon)
        if self.use_latent:
            for i in range(num_losses):
                metrics.update(all_metrics_latent[i])

        metrics["loss"] = loss
        return metrics


def perp_loss(yhat, y):
    """Implementation of the perpendicular loss.

    Args:
        yhat: Predicted reconstruction. Must be complex.
        y: Target reconstruction. Must be complex.

    Returns:
        Dict[str, scalar]:

    References:
        Terpstra, et al. "Rethinking complex image reconstruction:
        ⟂-loss for improved complex image reconstruction with deep learning."
        International Society of Magnetic Resonance in Medicine Annual Meeting
        2021.
    """
    if cplx.is_complex(yhat):
        yhat = torch.view_as_real(yhat)
    if cplx.is_complex(y):
        y = torch.view_as_real(y)

    P = torch.abs(yhat[..., 0] * y[..., 1] - yhat[..., 1] * y[..., 0]) / cplx.abs(y)
    l1 = torch.abs(cplx.abs(y) - cplx.abs(yhat))

    return {"p_perp_loss": torch.mean(P), "perp_loss": torch.mean(P + l1)}



"""Functions to register a MRCO-format dataset to the DatasetCatalog.
"""
import json
import logging
import os
import time

import h5py
import pandas as pd

from meddlr.data import DatasetCatalog, MetadataCatalog
from meddlr.utils import env

logger = logging.getLogger(__name__)
_PATH_MANAGER = env.get_path_manager()

__all__ = ["load_mrco_json", "register_mrco_scans"]


def load_mrco_json(json_file: str, image_root: str, dataset_name: str):
    """Load a json file with MRCO's scan annotation format.

    Currently supports reconstruction.

    Args:
        json_file (str): Full path to the json file in MRCO scan annotation
            format.
        image_root (str): The directory where the images in this json file
            exist.
        dataset_name (str): The name of the dataset
            (e.g. mridata_knee_2019_train).
        cipher_kwargs (Dict[str, Any]): Keyword arguments for loading metadata from
            cipher(s). See :func:`merge_ciphers_with_dataset_dicts`.

    Returns:
        List[Dict]: A list of dicts in meddlr standard format.

    Notes:
        1. This function does not read the image files.
           The results do not have the "kspace", "target", or "maps" fields.
    """
    json_file = _PATH_MANAGER.get_local_path(json_file)
    start_time = time.perf_counter()
    with open(json_file, "r") as f:
        data = json.load(f)
    logger.info(
        "Loading {} takes {:.2f} seconds".format(json_file, time.perf_counter() - start_time)
    )

    # TODO: Add any relevant metadata.
    dataset_dicts = []
    for d in data["images"]:
        dd = dict(d)
        if image_root is not None:
            file_name = _PATH_MANAGER.get_local_path(os.path.join(image_root, d["file_name"]))
        else:
            file_name = _PATH_MANAGER.get_local_path(d["file_path"])
        dd["file_name"] = file_name

        # TODO: Clean this up
        if any(dataset_name.startswith(x) for x in ["stanford_qDESS_knee_2020"]):
            with h5py.File(dd["file_name"], "r") as f:
                dd["kspace_shape"] = f["kspace"].shape

        dataset_dicts.append(dd)

    # Load metadata from metadata file.
    metadata_file = MetadataCatalog.get(dataset_name).metadata_file
    if metadata_file:
        dataset_dicts = _load_metadata(metadata_file, dataset_dicts)

    return dataset_dicts


def register_mrco_scans(name, metadata, json_file, image_root: str = None):
    """
    Register a dataset in COCO's json annotation format for
    instance detection, instance segmentation and keypoint detection.
    (i.e., Type 1 and 2 in http://cocodataset.org/#format-data.
    `instances*.json` and `person_keypoints*.json` in the dataset).

    This is an example of how to register a new dataset.
    You can do something similar to this function, to register new datasets.

    Args:
        name (str): the name that identifies a dataset, e.g. "coco_2014_train".
        metadata (dict): extra metadata associated with this dataset.  You can
            leave it as an empty dict.
        json_file (str): path to the json instance annotation file.
        image_root (str): directory which contains all the images.
    """
    # 1. register a function which returns dicts
    DatasetCatalog.register(name, lambda: load_mrco_json(json_file, image_root, name))

    # 2. Optionally, add metadata about this dataset,
    # since they might be useful in evaluation, visualization or logging
    MetadataCatalog.get(name).set(
        json_file=json_file, image_root=image_root, evaluator_type="coco", **metadata
    )


def _load_metadata(metadata_file, dataset_dicts):
    """Loads metadata in csv file into dataset dictionaries.

    Args:
        metadata_file (str): A csv file with metadata.
        dataset_dicts (List[Dict]): List of dataset dictionaries.

    Returns:
        dataset_dicts: The dataset dictionaries with "_metadata" key.
            This key maps to the dictionary of metadata.
    """
    metadata = pd.read_csv(_PATH_MANAGER.get_local_path(metadata_file))
    metadata = metadata.loc[:, ~metadata.columns.str.contains("^Unnamed")]

    for dd in dataset_dicts:
        dd_metadata = metadata[metadata["fname"] == os.path.basename(dd["file_name"])]
        assert len(dd_metadata) == 1
        dd_metadata = dd_metadata.to_dict("index")
        dd_metadata = dd_metadata[list(dd_metadata.keys())[0]]
        dd["_metadata"] = dd_metadata

    return dataset_dicts



import logging
import weakref
from bisect import bisect_right
from numbers import Number
from typing import Any, Callable, Dict, List, Mapping, Sequence, Tuple, Union

import numpy as np

from meddlr.transforms.param_kind import ParamKind
from meddlr.utils import env
from meddlr.utils.events import get_event_storage


class TFScheduler:
    """Abstract class that helps schedule parameters for different transforms.

    This class helps with scheduling parameters for classes which have schedulable
    parameters (:cls:`SchedulableMixin`). This may be helpful when estabilishing
    some curricula for augmentation parameters.

    Each :cls:`TFScheduler` (or scheduler) operates on some set of base parameters
    (``base_params``) available in the :cls:`SchedulableMixin` (or schedulable)
    class. These base parameters define the maximum values or operating ranges of
    the parameters that are to be used by the schedulable. See :cls:`SchedulableMixin`
    for more details.

    Scheduling requires access to the current iteration (or epoch) count.
    However, when these schedulers are used across different workers,
    the synchronized iteration count is not available. We build around this
    issue by maintaining a global and local iteration count, where the total
    iteration is approximately equal to the sum of the two. When worker states
    are synchronized with the main thread, the global iteration count is updated
    and the local count is reset to 0.

    Such a scheduling system requires a local iteration counter that is initialized and
    kept at 0 on the main thread. This state is computed from ``self._step``.
    When a process is forked, ``self._step == 0``. The state is routinely updated on
    each worker until the worker syncs with the state on the main thread. Upon sync,
    ``self._step`` is reset to 0, but the global iteration counter is updated. Thus,
    the sum ``global_iter + iter_fn(self._step)`` approximately provides the current
    iteration. See :meth:`get_iteration` for more detail.

    Attributes:
        tfm (weakref.proxy[ScheduleableMixin]): A weak reference to input ``tfm``.

    Notation:
        - ``t``: The estimated iteration.

    Note:
        Functionality has not been tested with MacOS or Windows OS. It may work only with
        forking (default on Unix) not spawning (default on Windows and MacOS).
    """

    def __init__(
        self, tfm, params: Union[str, List[str], Tuple[str]] = None, iter_fn: Callable = None
    ):
        """
        Args:
            tfm (SchedulableMixin): An instance of :cls:`SchedulableMixin`.
                Typically a transform.
            params (List[str]): Parameter name(s) for ``tfm`` that should be scheduled.
                These parameters should be accessible from ``tfm._params``.
            iter_fn (Callable): A function that returns the current iteration.
                This function should take in one argument, ``self._step``.
                Only used when the scheduler is being used across multiple workers
                (e.g. data loading).
        """
        if not env.is_main_process():
            raise RuntimeError(f"{type(self)} must be constructed on the main thread.")

        self.tfm: SchedulableMixin = weakref.proxy(tfm)
        self._params = []
        self._register_parameters(params)
        self._iter_fn = iter_fn
        self._step = 0

    def get_params(self) -> Dict[str, Any]:
        """Returns dictionary of ``self.tfm`` parameters after scheduling.

        Returns:
            Dict[str, Any]: The parameters after being passed through the scheduler.
        """
        names = self._parameter_names()
        params = {}
        for pname in names:
            kind = self.tfm._param_kinds.get(pname, ParamKind.SINGLE_ARG)
            self._fill_param_dict(
                params, pname, self._compute_value(self._get_tfm_param_val(pname), kind)
            )
        return params

    def get_iteration(self) -> int:
        """Returns the estimated iteration.

        Returns:
            int: The estimated iteration.
        """
        if env.is_main_process():
            return get_event_storage().iter

        # The EventStorage object is not always synchronized betweeen different
        # forks. Therefore, directly accessing get_event_storage().iter may not
        # give the correct iteration estimate. `self._step` keeps track of changes
        # in iterations for each worker based on the computation from `self._iter_fn`.
        # `self._step` is and should always be 0 on the main process. In other words,
        # it is only updated on each worker and is reset to 0 everytime states are
        # synchronized.
        # Therefore, the true iteration count can be estimated by adding the
        # iteration number from EventStorage with the estimated elapsed iteration
        # between synchronization periods given by `self._iter_fn(self._step)`.
        base_iter = get_event_storage().iter
        delta_iter = self._iter_fn(self._step)
        return base_iter + delta_iter

    def step(self, n: int = 1):
        """Take ``n`` step(s).

        This function is used to maintain the step count
        used to compute the current iteration number.
        ``n`` should be configured to be compatible with ``self._iter_fn``.
        If you are using :func:`build_iter_func`, ``n`` should correspond
        to the number of examples in the batch.

        This function does not do anything if ``self`` is executing on the
        main thread.

        Args:
            n (int): The number of steps/examples to increase internal
                step count by.
        """
        if env.is_main_process():
            return

        self._step += n

    def _compute_value(self, value, param_kind: ParamKind):
        raise NotImplementedError

    def _parameter_names(self) -> List[str]:
        return self._params if self._params is not None else list(self.tfm.params().keys())

    def _get_tfm_keys(self):
        def _dfs_pname(params, prefix=""):
            out = []
            if isinstance(params, Mapping):
                for k, v in params.items():
                    out.extend(_dfs_pname(v, f"{prefix}.{k}"))
            else:
                out.append(prefix)
            return out

        out = _dfs_pname(self.tfm._params)
        return [x[1:] if x.startswith(".") else x for x in out]

    def _register_parameters(self, names: Union[str, Sequence[str]]):
        # TODO (arjundd): names=None does not work.
        if isinstance(names, str):
            names = [names]
        names = [x for x in self._get_tfm_keys() if any(x.startswith(n) for n in names)]
        unknown_params = set(names) - set(self._get_tfm_keys())
        if len(unknown_params) > 0:
            raise ValueError(
                f"Unknown parameters for transform {self.tfm.__class__.__name__}: {unknown_params}"
            )
        params = self._parameter_names()
        params = [n for n in names if n not in params]
        self._params.extend(params)

    def _unregister_parameters(self, names: Union[str, Sequence[str]]):
        if isinstance(names, str):
            names = [names]
        names = [x for x in self._get_tfm_keys() if any(x.startswith(n) for n in names)]
        params = self._parameter_names()
        params = [x for x in params if x not in names]
        self._params = params

    def _get_tfm_param_val(self, pname: str):
        pname = _parse_pname(pname)
        if isinstance(pname, str):
            return self.tfm._params[pname]

        pval = self.tfm._params
        for pn in pname:
            pval = pval[pn]
        return pval

    def _fill_param_dict(self, params: Dict, pname: str, pval):
        pname = _parse_pname(pname)
        if isinstance(pname, str):
            params[pname] = pval
            return

        for idx, pn in enumerate(pname):
            if pn not in params:
                params[pn] = pval if idx == len(pname) - 1 else {}
            params = params[pn]

    def _repr_args(self) -> List[str]:
        return ["tfm", "_params"]

    def __repr__(self) -> str:
        args = self._repr_args()
        args = {k: getattr(self, k) for k in args}
        args_str = "\n  ".join(
            [f"{k}={v.__repr__() if hasattr(v, '__repr__') else v}," for k, v in args.items()]
        )
        return "{}(\n  {}\n)".format(type(self).__name__, args_str)

    def __str__(self) -> str:
        return self.__repr__()


class SchedulableMixin:
    """Mixin for any class that should be compatible with :cls:`TFScheduler`."""

    _params: Dict[str, Union[Number, Tuple[Number, Number]]]
    _param_kinds: Dict[str, ParamKind]
    _schedulers: List[TFScheduler]

    def base_params(self):
        """The base that defines the maximum value or range for different parameters.

        Currently, this only supports numeric parameter types.

        Returns:
            Dict[str, Any]: The maximum value or range of different parameters.
        """
        return self._params

    def validate_schedulers(self) -> None:
        """Verifies that schedulers are valid.

        A valid set of schedulers must fulfill at least the following criteria:
            1. Any two different schedulers should not operate on any of the same paramters.
            2. All schedulers must have a weak reference to this schedulable.

        Raises:
            ValueError: If parameters overlap between any two schedulers
                or schedulers do not have a weakref to this schedulable.
        """
        schedulers = self._schedulers

        # No parameters should overlap between schedulers.
        all_params = [set(scheduler._params) for scheduler in schedulers]
        for i in range(len(all_params)):
            for j in range(i + 1, len(all_params)):
                if len(all_params[i] & all_params[j]) > 0:
                    raise ValueError(f"Parameters overlapping between schedulers {i} and {j}")

        # All schedulers should be tied to this schedulable.
        if any(s.tfm != self.__weakref__ for s in self._schedulers):
            raise ValueError(
                "Schedulers can only be configured for the transform it is registered to"
            )

    def register_schedulers(
        self, schedulers: Sequence[TFScheduler], overwrite_params: bool = False
    ) -> None:
        """Register scheduler(s).

        Args:
            schedulers (Sequence[TFScheduler]): The sequence of schedulers to
                register. These schedulers will be used to operate on certain
                parameters.
            overwrite_params (bool, optional): If ``True``, more recently added
                schedulers will have exclusive priority to manage parameter values.
                For example, if schedulers ``A`` and ``B`` were to be added in that
                order, and they were both configured to operate on some parameter
                ``param1``, then ``B``'s scheduling would be used, as it is the most
                recent scheduler.
        """
        if isinstance(schedulers, TFScheduler):
            schedulers = [schedulers]
        if overwrite_params:
            new_params = [name for s in schedulers for name in s._parameter_names()]
            for s in self._schedulers:
                s._unregister_parameters([p for p in new_params if p in s._parameter_names()])

        self._schedulers.extend(schedulers)
        self.validate_schedulers()

    def schedulers(self) -> List[TFScheduler]:
        return self._schedulers


class WarmupTF(TFScheduler):
    """A warmup scheduler.

    In warmup, a numeric value is scaled between some lower (:math:`b_l`)
    and upper (:math:`b_u`) bound over some number of iterations (i.e.
    the ``warmup_iters``). If ``t >= warmup_iters``, the value is clamped
    at the maximum value.

    If the base parameter is a scalar value, that value is assumed to be the
    upper bound with a default lower bound of ``0``. If the parameter is a
    tuple, it is assumed to be the semi-open range corresponding to
    [:math:`b_l`, :math:`b_u`).

    For notation, see cls:`TFScheduler`.
    """

    def __init__(
        self,
        tfm: SchedulableMixin,
        warmup_iters: int,
        warmup_method: str = "linear",
        delay_iters: int = 0,
        gamma: float = 1.0,
        params=None,
    ):
        """
        Args:
            tfm: See :cls:`TFScheduler`.
            warmup_iters (int): Number of iterations to warmup.
            warmup_method (str, optional): One of [``'linear'``, ``'exp'``].
            delay_iters (int, optional): Number of iterations to delay scheduling by.
            gamma (float, optional): Scaling factor for time constant in exponential (``'exp'``)
                scheduling.
            params: See :cls:`TFScheduler`.
        """
        _logger = logging.getLogger(__name__)

        if warmup_method not in ("linear", "exp"):
            raise ValueError(
                f"`warmup_method` must be one of (None, 'linear', 'exp'). Got '{warmup_method}'"
            )
        if gamma < 0:
            raise ValueError("gamma must be >=0")
        if warmup_iters == 0 and delay_iters == 0:
            _logger.warning(
                "No warmup or delay time specified. " "This will functionally be a no-op scheduler."
            )

        self.delay_iters = delay_iters
        self.warmup_iters = warmup_iters
        self.warmup_method = warmup_method
        self.gamma = gamma

        super().__init__(tfm, params)

    def _compute_value(self, value, kind: ParamKind):
        if kind == ParamKind.MULTI_ARG and isinstance(value, (list, tuple)):
            return [self._compute_value(v, kind.SINGLE_ARG) for v in value]

        is_number = isinstance(value, Number)
        if is_number:
            value = (0, value)
        else:
            assert isinstance(value, Sequence) and len(value) == 2

        alpha = _get_warmup_factor_at_iter(
            self.warmup_method,
            self.get_iteration(),
            self.warmup_iters,
            self.delay_iters,
            self.gamma,
        )
        lb, ub = tuple(value)
        ub = lb + alpha * (ub - lb)
        return ub if is_number else (lb, ub)

    def _repr_args(self) -> List[str]:
        base = super()._repr_args()
        base.extend(["warmup_iters", "warmup_method", "delay_iters", "gamma"])
        return base


class WarmupMultiStepTF(TFScheduler):
    """Multi-step warmup scheduler.

    This function works exactly like :cls:`WarmupTF` with the difference that
    all updates (or scheduling steps) are taken at discrete step times (or
    *milestones*). After the final milestone, the value is clamped at the upper
    bound for that parameter.

    This class is useful for epoch-level scheduling.
    """

    def __init__(
        self,
        tfm: SchedulableMixin,
        warmup_milestones: Sequence[int],
        warmup_method: str = "linear",
        gamma: float = 1.0,
        params=None,
    ):
        """
        Args:
            tfm: See :cls:`TFScheduler`.
            warmup_milestones (Sequence[int]): Discrete iterations at which to take a
                step on the scheduler.
            warmup_method (str, optional): One of [``'linear'``, ``'exp'``].
            gamma (float, optional): Scaling factor for time constant in exponential (``'exp'``)
                scheduling.
            params: See :cls:`TFScheduler`.
        """
        if warmup_method not in (None, "linear", "exp"):
            raise ValueError(
                f"`warmup_method` must be one of (None, 'linear', 'exp'). Got '{warmup_method}'"
            )
        if gamma < 0:
            raise ValueError("gamma must be >=0")

        self.warmup_milestones = warmup_milestones
        self.warmup_method = warmup_method
        self.gamma = gamma

        super().__init__(tfm, params)

    def _compute_value(self, value, kind: ParamKind):
        if kind == ParamKind.MULTI_ARG and isinstance(value, (list, tuple)):
            return [self._compute_value(v, kind.SINGLE_ARG) for v in value]

        t = self.get_iteration()
        step = bisect_right(self.warmup_milestones, t)
        total_steps = len(self.warmup_milestones)

        is_number = isinstance(value, Number)
        if is_number:
            value = (0, value)
        else:
            assert isinstance(value, Sequence) and len(value) == 2

        alpha = _get_warmup_factor_at_iter(self.warmup_method, step, total_steps, 0, self.gamma)
        alpha = (step > 0) * alpha

        lb, ub = tuple(value)
        ub = lb + alpha * (ub - lb)
        return ub if is_number else (lb, ub)

    def _repr_args(self) -> List[str]:
        base = super()._repr_args()
        base.extend(["warmup_milestones", "warmup_method", "gamma"])
        return base


def _get_warmup_factor_at_iter(
    method: str, iter: int, warmup_iters: int, delay_iters: int, gamma: float = None
) -> float:
    """
    TODO: Merge this with the lr scheduler logic

    Args:
        method (str): warmup method; either "constant" or "linear".
        iter (int): iteration at which to calculate the warmup factor.
        warmup_iters (int): the number of warmup iterations.
        warmup_factor (float): the base warmup factor (the meaning changes
            according to the method used).

    Returns:
        float: the effective warmup factor at the given iteration.
    """
    if iter >= warmup_iters:
        return 1.0
    if iter < delay_iters:
        return 0.0
    assert warmup_iters > delay_iters

    if method == "constant":
        return 1.0 / (warmup_iters - delay_iters)
    elif method == "linear":
        return min((iter - delay_iters) / (warmup_iters - delay_iters), 1.0)
    elif method == "exp":
        assert gamma is not None
        tau = (warmup_iters - delay_iters) / gamma
        return min((1 - np.exp(-(iter - delay_iters) / tau)) / (1 - np.exp(-gamma)), 1.0)
    else:
        raise ValueError("Unknown warmup method: {}".format(method))


def _parse_pname(pname: str):
    if "." not in pname:
        return pname
    return pname.split(".")



from typing import Tuple

import torch

from meddlr.ops import complex as cplx
from meddlr.transforms.build import TRANSFORM_REGISTRY
from meddlr.transforms.transform import Transform


@TRANSFORM_REGISTRY.register()
class NoiseTransform(Transform):
    """A model that adds additive white noise."""

    def __init__(
        self,
        std_dev: int = None,
        use_mask: bool = True,
        rho: float = None,
        seed: int = None,
        generator: torch.Generator = None,
    ):
        if seed is None and generator is None:
            raise ValueError("One of `seed` or `generator` must be specified.")
        self.std_dev = std_dev
        self.use_mask = use_mask
        self.rho = rho
        self.seed = seed

        gen_state = None
        if generator is not None:
            gen_state = generator.get_state()
        self._generator_state = gen_state

    def apply_kspace(self, kspace: torch.Tensor):
        return self._add_noise(kspace)

    def _generator(self, data):
        seed = self.seed

        g = torch.Generator(device=data.device)
        if seed is None:
            g.set_state(self._generator_state)
        else:
            g = g.manual_seed(seed)
        return g

    def _add_noise(self, data: torch.Tensor):
        noise_std = self.std_dev
        subsample_masks = self.rho is not None and self.rho != 1

        mask = None
        if self.use_mask:
            mask = cplx.get_mask(data)
        if mask is None and subsample_masks:
            mask = torch.ones(data.shape)

        g = self._generator(data)
        if cplx.is_complex(data):
            noise = noise_std * torch.randn(data.shape + (2,), generator=g, device=data.device)
            noise = torch.view_as_complex(noise)
        else:
            noise = noise_std * torch.randn(data.shape, generator=g, device=data.device)

        if self.rho is not None and self.rho != 1:
            mask = self.subsample_mask(mask)
        if mask is not None:
            noise = noise * mask
        aug_kspace = data + noise

        return aug_kspace

    def subsample_mask(self, mask: torch.Tensor, generator=None):
        """Subsamples mask to add the noise to.

        Currently done uniformly at random.
        """
        rho = self.rho
        shape = mask.shape
        mask = mask.view(-1)

        # TODO: this doesnt work if the matrix is > 2*24 in size.
        # TODO: make this a bit more optimized
        num_valid = torch.sum(mask)
        weights = mask / num_valid
        num_samples = int((1 - rho) * num_valid)
        samples = torch.multinomial(weights, num_samples, replacement=False, generator=generator)
        mask[samples] = 0

        mask = mask.view(shape)
        return mask

    @classmethod
    def from_cfg(cls, cfg, seed=None, **kwargs):
        cfg = cfg.MODEL.CONSISTENCY.AUG.NOISE
        return cls(cfg.STD_DEV, scheduler=cfg.SCHEDULER, mask=cfg.MASK, seed=seed, **kwargs)

    def _eq_attrs(self) -> Tuple[str]:
        return ("std_dev", "use_mask", "rho", "seed", "_generator_state")
